/*
 *  pv_myinit_x11.c (for X11)
 *  ProteinViewer
 *
 *  Created by Naoto Morikawa on 6/15/06.
 *  Copyright 2006 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 */

#include <stdlib.h>
#include <GL/glut.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "pv_main.h"
#include "pv_read_data.h"
#include "pv_primitives.h"

extern void print_trn_table(void);    /* trn_table.c */
extern void print_npath_table(void);
extern void setup_transition_table(float ave_dis);
extern int setup_npath_table(void);
extern void create_display_list(void);
extern struct pe_slt_tile * read_data(char ** argv, struct pe_atom * atoms,
				      struct pe_coor_system * pt_coor);
extern void set_rotation_matrix(struct pe_slt_tile * slt_tiles);

extern struct pv_macro my_macro;
extern struct pe_coor_system  my_coor;
extern struct pe_settings my_settings;
extern struct pe_trn_table my_trn_table[];

float Mtile_coor[16]; /* 4 x 4 matrix */
float Tcenter[3]; /* Translation */
float Tcenter_save[3];
float Ssize [4]; /* Scale + max */

GLfloat light_diffuse[]={0.5F, 0.5F, 0.5F, 1.0F}; 
GLfloat light_specular[]={1.0F, 1.0F, 1.0F, 1.0F}; 
GLfloat light_ambient[]={0.0F, 0.0F, 0.0F, 1.0F};
GLfloat light_position[]={1.0F, 0.0F, 0.0F, 0.0F};
GLfloat light_position2[]={0.0F, 1.0F, 0.0F, 0.0F};
GLfloat light_position3[]={0.0F, 1.0F, 1.0F, 0.0F};
GLfloat lmodel_ambient[]={0.2F, 0.2F, 0.2F, 1.0F};

void print_macro(struct pv_macro *pt_strct)
{
	int i;
	
	fprintf(STDOUT,"\n");
	fprintf(STDOUT, "macro: mod |mol |chn |seq |res |atm |\n");
	fprintf(STDOUT, "macro: %4d|%4d|%4d|%4d|%4d|%4d|\n", pt_strct->mod_num,
		pt_strct->mol_num, pt_strct->chn_num, pt_strct->seq_num,
		pt_strct->res_num, pt_strct->atm_num);	
	if ( pt_strct->model != NULL ) {
		fprintf(STDOUT, "macro: <models>\n");
		fprintf(STDOUT, "macro:            [seq_head, seq_tail]\n");
		for (i = 1; i <= pt_strct->mod_num; i++) {
			fprintf(STDOUT, "macro: model[%2d]= [ %7d,  %7d]\n", i, 
				pt_strct->model[i].seq_head,
				pt_strct->model[i].seq_tail);
		}
	}
	
	fprintf(STDOUT, "macro: <molecules>\n");
	fprintf(STDOUT, "macro:                [chn_head, chn_tail] id name\n");
	for (i = 1; i <= pt_strct->mol_num; i++) {
		fprintf(STDOUT, "macro: molecules[%2d]= [ %7d,  %7d] %2d %s\n",
			i, pt_strct->molecule[i].chain_head,
			pt_strct->molecule[i].chain_tail,
			pt_strct->molecule[i].id,
			pt_strct->molecule[i].name);
	}
	
	fprintf(STDOUT, "macro: <chains>\n");	
	fprintf(STDOUT,
		"macro:            [atm_head, atm_tail] chn_id |mol_id |\n");
	for (i = 1; i <= pt_strct->chn_num; i++) {
		fprintf(STDOUT, "macro: chain[%2d]= [ %7d,  %7d] %7c|%7d|\n", i,
			pt_strct->chain[i].atom_head,
			pt_strct->chain[i].atom_tail, pt_strct->chain[i].id,
			pt_strct->chain[i].mol_id);		
	}
	
	if ( pt_strct->sequence != NULL ) {
		fprintf(STDOUT, "macro: <sequences>\n");
		fprintf(STDOUT,
"macro:                [atm_head, atm_tail] chn_id |mol_id |mod_num [tile_head,tile_tail]\n");
		for (i = 1; i <= pt_strct->seq_num; i++) {
			fprintf(STDOUT,
"macro: sequence[%3d]= [ %7d,  %7d] %7c|%7d|%7d [ %8d, %8d]\n",
				i, pt_strct->sequence[i].atom_head,
				pt_strct->sequence[i].atom_tail,
				pt_strct->sequence[i].id,
				pt_strct->sequence[i].mol_id,
				pt_strct->sequence[i].mod_num,
				pt_strct->sequence[i].tile_head,
				pt_strct->sequence[i].tile_tail);
		}
	}
	
	fprintf(STDOUT, "macro: <settings>\n");
	fprintf(STDOUT, "macro: tiles_per_res = %4.1f\n",
		my_macro.tiles_per_res);
	fprintf(STDOUT, "macro: along_backbone = %d\n",
		my_macro.along_backbone);
	
	fprintf(STDOUT, "macro: end of my_macro \n");
	fprintf(STDOUT, "\n");
	fflush(STDOUT);
	return;
}

void print_atoms(struct pe_atom *atoms)
{
	int mod_ct, seq_ct, atm_ct;
	
	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "[PDB data] (#atoms= %d)\n", my_macro.atm_num);
	fprintf(STDOUT, "#sequences= %d, #molecules= %d, #models= %d\n",
		my_macro.seq_num, my_macro.mol_num, my_macro.mod_num);
	
	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for ( seq_ct= my_macro.model[mod_ct].seq_head;
		      seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			fprintf(STDOUT, "\n");
			fprintf(STDOUT,
"Atom sequence %d (from %d-th atom to %d-th atom)\n",
				seq_ct, my_macro.sequence[seq_ct].atom_head,
				my_macro.sequence[seq_ct].atom_tail);
			fprintf(STDOUT,
"       [  No. ]  ser.  name res chn num  (       x,        y,        z) occu. id\n");
			for (atm_ct = my_macro.sequence[seq_ct].atom_head;
			     atm_ct <= my_macro.sequence[seq_ct].atom_tail;
			     atm_ct++) {  
				fprintf(STDOUT,"atom : [%6d] %6d ",
					atm_ct,
					my_macro.atom[atm_ct].serial_num);
				if (my_macro.atom[atm_ct].atom_name == N_CC)
					fprintf(STDOUT," C  ");
				else if (my_macro.atom[atm_ct].atom_name == N_CA)
					fprintf(STDOUT," CA ");
				else if (my_macro.atom[atm_ct].atom_name == N_NN)
					fprintf(STDOUT," N  ");
				else if (my_macro.atom[atm_ct].atom_name == N_CR)
					fprintf(STDOUT," CR ");
				else if (my_macro.atom[atm_ct].atom_name == N_CY)
					fprintf(STDOUT," CY ");
				else if (my_macro.atom[atm_ct].atom_name == N_CW)
					fprintf(STDOUT," CW ");
				else
					fprintf(STDOUT," UN ");
				
				fprintf(STDOUT," %.3s  %c  %4d (%8.3f, %8.3f, %8.3f) %4.3f %.4s\n",
					my_macro.atom[atm_ct].res_name,
					my_macro.atom[atm_ct].chain_id,
					my_macro.atom[atm_ct].res_seq,
					my_macro.atom[atm_ct].x,
					my_macro.atom[atm_ct].y,
					my_macro.atom[atm_ct].z,
					my_macro.atom[atm_ct].occupancy,
					my_macro.atom[atm_ct].seg_id);
			}
			fprintf(STDOUT,"End of atom sequence\n");
		}
		fprintf(STDOUT,"End of model\n");
	}
	fprintf(STDOUT,"End of [PDB data]\n");
	fprintf(STDOUT,"\n");
	fflush(STDOUT);
	return;
}

void print_coordinates(struct pe_coor_system *pt_coor)
{
	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "[Coordinates]:  Origin = ( %8.3f, %8.3f, %8.3f ) \n",
		pt_coor->origin.x, pt_coor->origin.y, pt_coor->origin.z );
	fprintf(STDOUT, "[Coordinates]:  x-axis = ( %8.3f, %8.3f, %8.3f ) \n",
		pt_coor->axes[0].x, pt_coor->axes[0].y, pt_coor->axes[0].z);
	fprintf(STDOUT, "[Coordinates]:  y-axis = ( %8.3f, %8.3f, %8.3f ) \n",
		pt_coor->axes[1].x, pt_coor->axes[1].y, pt_coor->axes[1].z);
	fprintf(STDOUT, "[Coordinates]:  z-axis = ( %8.3f, %8.3f, %8.3f ) \n",
		pt_coor->axes[2].x, pt_coor->axes[2].y, pt_coor->axes[2].z);
	fprintf(STDOUT, "[Coordinates]:  w-axis = ( %8.3f, %8.3f, %8.3f ) \n",
		pt_coor->axes[3].x, pt_coor->axes[3].y, pt_coor->axes[3].z);
	return;
}

void print_tiles(struct pe_slt_tile *s_tiles)
{
	int i, seq_ct;
	int atm_ind;
	
	seq_ct = 1;
	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "[Slant tiles] (#tiles= %d)\n", my_macro.tile_num);
	fprintf(STDOUT, "Tile sequence %d\n", seq_ct);
	
	for (i = 1; i <= my_macro.tile_num; i++) {
		fprintf(STDOUT, "tile : [%d] x^%d y^%d z^%d w^%d %s,",
			i, s_tiles[i].top[0], s_tiles[i].top[1],
			s_tiles[i].top[2], s_tiles[i].top[3],
			my_trn_table[s_tiles[i].type].str);
		if (s_tiles[i].code == C_UP)
			fprintf(STDOUT, " code= U,");
		else
			fprintf(STDOUT, " code= D,");
		atm_ind = s_tiles[i].error_atom;
		fprintf(STDOUT, " atom= %d", s_tiles[i].error_atom);
		if (s_tiles[i].error_atom > 0) {
			if (my_macro.atom[atm_ind].atom_name == N_CC)
				fprintf(STDOUT, " (C )\n");
			else if (my_macro.atom[atm_ind].atom_name == N_CA)
				fprintf(STDOUT, " (CA)\n");
			else if (my_macro.atom[atm_ind].atom_name == N_NN)
				fprintf(STDOUT, " (N )\n");
			else if (my_macro.atom[atm_ind].atom_name == N_CR)
				fprintf(STDOUT, " (CR)\n");
			else if (my_macro.atom[atm_ind].atom_name == N_CY)
				fprintf(STDOUT, " (CY)\n");
			else if (my_macro.atom[atm_ind].atom_name == N_CW)
				fprintf(STDOUT, " (CW)\n");
			else
				fprintf(STDOUT, " (UN)\n");
		} else {
			fprintf(STDOUT, " (UN)\n");
		}
		if (DEBUG) {
			fprintf(STDOUT, "       center= (%8.3f, %8.3f, %8.3f), ",
				s_tiles[i].cx,s_tiles[i].cy,s_tiles[i].cz);
			fprintf(STDOUT, "error= %f against atom [%d]\n",
				s_tiles[i].error, s_tiles[i].error_atom);
			if (my_settings.rotate_tile == PV_TRUE)
				fprintf(STDOUT,
					"       direction= (%8.3f, %8.3f, %8.3f)\n",
					s_tiles[i].dx, s_tiles[i].dy,
					s_tiles[i].dz);
		}
		fflush(STDOUT);
		
		if ((s_tiles[i].ter > 0) && (i < my_macro.tile_num)) {
			seq_ct++;
			fprintf(STDOUT, "End of tile sequence\n");
			fprintf(STDOUT, "\n");
			fprintf(STDOUT, "Tile sequence %d\n", seq_ct);
		}
	}
	fprintf(STDOUT,"End of tile sequence\n");
	fprintf(STDOUT,"End of [Slant tiles]\n");
	fprintf(STDOUT,"\n");
	fflush(STDOUT);
	return;
}

void print_t_vec(struct pe_vector *t_vec)
{
	int mod_ct, seq_ct, t_ct;
	
	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "[t_vec] (#tiles= %d)\n", my_macro.tile_num);
	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for ( seq_ct = my_macro.model[mod_ct].seq_head;
		      seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			fprintf(STDOUT, "Vector sequence %d\n", seq_ct);
			fprintf(STDOUT,
				"         [  No. ]  (       x,        y,        z)  length\n");
			for (t_ct = my_macro.sequence[seq_ct].tile_head;
			     t_ct <= my_macro.sequence[seq_ct].tile_tail;
			     t_ct++) {
				fprintf(STDOUT, "t_vec  : [%6d] ", t_ct);
				
				fprintf(STDOUT,
					" (%8.3f, %8.3f, %8.3f) %8.3f\n",
					t_vec[t_ct].x, t_vec[t_ct].y,
					t_vec[t_ct].z,
					t_vec[t_ct].length);
			}
			fprintf(STDOUT,"End of vector sequence\n");
		}
		fprintf(STDOUT,"End of model\n");
	}
	fprintf(STDOUT,"End of [t_vec] vector\n");
	fprintf(STDOUT,"\n");
	fflush(STDOUT);
	return;
}

void print_vectors(struct pe_vector *atm_vec)
{
	int mod_ct, seq_ct, atm_ct;
	
	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "[PDB vector] (#atoms= %d)\n", my_macro.atm_num);
	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for ( seq_ct = my_macro.model[mod_ct].seq_head;
		      seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			fprintf(STDOUT, "Vector sequence %d\n", seq_ct);
			fprintf(STDOUT,
"         [  No. ] name (       x,        y,        z)  length\n");
			for (atm_ct = my_macro.sequence[seq_ct].atom_head;
			     atm_ct <= my_macro.sequence[seq_ct].atom_tail;
			     atm_ct++) {  
				fprintf(STDOUT, "vector : [%6d] ", atm_ct);
				if (my_macro.atom[atm_ct].atom_name == N_CC)
					fprintf(STDOUT, " C  ");
				else if (my_macro.atom[atm_ct].atom_name
					 == N_CA)
					fprintf(STDOUT, " CA ");
				else if (my_macro.atom[atm_ct].atom_name
					 == N_NN)
					fprintf(STDOUT, " N  ");
				else if (my_macro.atom[atm_ct].atom_name
					 == N_CR)
					fprintf(STDOUT, " CR ");
				else if (my_macro.atom[atm_ct].atom_name
					 == N_CY)
					fprintf(STDOUT, " CY ");
				else if (my_macro.atom[atm_ct].atom_name
					 == N_CW)
					fprintf(STDOUT, " CW ");
				else
					fprintf(STDOUT, " UN ");
				
				fprintf(STDOUT,
					" (%8.3f, %8.3f, %8.3f) %8.3f\n",
					atm_vec[atm_ct].x, atm_vec[atm_ct].y,
					atm_vec[atm_ct].z,
					atm_vec[atm_ct].length);
			}
			fprintf(STDOUT,"End of vector sequence\n");
		}
		fprintf(STDOUT,"End of model\n");
	}
	fprintf(STDOUT,"End of [PDB vector]\n");
	fprintf(STDOUT,"\n");
	fflush(STDOUT);
	return;
}

void init_macro(void)
{
	int i;
	
	my_macro.mod_num = 0;
	my_macro.model = NULL;
	my_macro.mol_num = 0;
	my_macro.chn_num = 0;
	my_macro.res_num = 0;
	my_macro.atom = NULL;
	my_macro.seq_num = 0;
	my_macro.sequence = NULL;
	
	my_macro.tile_max_num = 0;
	my_macro.tile_num = 0;
	my_macro.tiles_per_res = 0.0;
	my_macro.dist_factor = 0.0;
	my_macro.rotate_tile = PV_FALSE;
	
	for (i = 0; i < MOL_SIZE; i++) {
		my_macro.molecule[i].chain_head = 0;
		my_macro.molecule[i].chain_tail = 0;
		my_macro.molecule[i].id = 0;
		strcpy(my_macro.molecule[i].name, "");
	}
	
	for (i = 0; i < CHN_SIZE; i++) {
		my_macro.chain[i].atom_head = 0;
		my_macro.chain[i].atom_tail = 0;
		my_macro.chain[i].id = '-';
		my_macro.chain[i].mol_id = 0;
		my_macro.chain[i].mod_num = 0;
		my_macro.chain[i].tile_head = 0;
		my_macro.chain[i].tile_tail = 0;
	}
	return;
}

void init_coor(struct pe_coor_system * pt_coor)
{
	int i;
	
	for (i = 0; i < 4; i++) {
		pt_coor->axes[i].x = 0.0;
		pt_coor->axes[i].y = 0.0;
		pt_coor->axes[i].z = 0.0;
		pt_coor->axes[i].length = 0.0;
	}
	pt_coor->origin.x = 0.0;
	pt_coor->origin.y = 0.0;
	pt_coor->origin.z = 0.0;
	pt_coor->origin.length = 0.0;
	pt_coor->ave_dis = 0.0;
	return;
}

int compute_atom(int seq_ct, struct pe_vector * atm_vec)
{
	int atm_ct;
	float wrk_x, wrk_y, wrk_z;
	
	for (atm_ct = my_macro.sequence[seq_ct].atom_head;
	     atm_ct < my_macro.sequence[seq_ct].atom_tail; atm_ct++) {
		wrk_x = my_macro.atom[atm_ct + 1].x - my_macro.atom[atm_ct].x;
		wrk_y = my_macro.atom[atm_ct + 1].y - my_macro.atom[atm_ct].y;
		wrk_z = my_macro.atom[atm_ct + 1].z - my_macro.atom[atm_ct].z;
		atm_vec[atm_ct].x = wrk_x;
		atm_vec[atm_ct].y = wrk_y;
		atm_vec[atm_ct].z = wrk_z;
		atm_vec[atm_ct].length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
					      + wrk_z * wrk_z);
		if (atm_vec[atm_ct].length == 0.0) {
			fprintf(STDERR,
				"compute_atom() : ERROR atm_vec[%d] == (0,0,0)\n",
				atm_ct);
			return FAIL;
		}
		if (0) fprintf(STDOUT,
			       "set_atom_vectors(): atm_vec[%d]= (%8.3f, %8.3f, %8.3f) %8.3f\n",
			       atm_ct, atm_vec[atm_ct].x, atm_vec[atm_ct].y,
			       atm_vec[atm_ct].z, atm_vec[atm_ct].length);
	}
	if (atm_ct < my_macro.atm_num) {
		wrk_x = my_macro.atom[atm_ct + 1].x - my_macro.atom[atm_ct].x;
		wrk_y = my_macro.atom[atm_ct + 1].y - my_macro.atom[atm_ct].y;
		wrk_z = my_macro.atom[atm_ct + 1].z - my_macro.atom[atm_ct].z;
		atm_vec[atm_ct].x = wrk_x;
		atm_vec[atm_ct].y = wrk_y;
		atm_vec[atm_ct].z = wrk_z;
		atm_vec[atm_ct].length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
					      + wrk_z * wrk_z);
		if (atm_vec[atm_ct].length == 0.0) {
			fprintf(STDERR,
				"compute_atom() : ERROR atm_vec[%d] == (0,0,0)\n",
				atm_ct);
			return FAIL;
		}
	} else {
		atm_vec[atm_ct].x = 0.0;
		atm_vec[atm_ct].y = 0.0;
		atm_vec[atm_ct].z = 0.0;
		atm_vec[atm_ct].length = 0.0;
	}
	return SUCCESS;
}

int check_atom_position(struct pe_vector * atm_vec)
{
	int mod_ct, seq_ct, atm_ct;
	float wrk_x, wrk_y, wrk_z;
	
	wrk_x = my_macro.atom[1].x;
	wrk_y = my_macro.atom[1].y;
	wrk_z = my_macro.atom[1].z;
	for (mod_ct=1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for ( seq_ct= my_macro.model[mod_ct].seq_head;
		      seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			for (atm_ct = my_macro.sequence[seq_ct].atom_head;
			     atm_ct < my_macro.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				wrk_x = atm_vec[atm_ct].x + wrk_x;
				wrk_y = atm_vec[atm_ct].y + wrk_y;
				wrk_z = atm_vec[atm_ct].z + wrk_z;
			}
			if (FLOAT_NEQ(my_macro.atom[atm_ct].x - wrk_x)
			    || FLOAT_NEQ(my_macro.atom[atm_ct].y - wrk_y)
			    || FLOAT_NEQ(my_macro.atom[atm_ct].z - wrk_z) ) {
				fprintf(STDERR,
"check_atom_position() : Error atom[%d].c* = (%8.3f, %8.3f, %8.3f) != (%8.3f, %8.3f, %8.3f)\n",
					atm_ct, my_macro.atom[atm_ct].x,
					my_macro.atom[atm_ct].y,
					my_macro.atom[atm_ct].z, wrk_x, wrk_y,
					wrk_z);
					return FAIL;
				}
				wrk_x = atm_vec[atm_ct].x + wrk_x;
				wrk_y = atm_vec[atm_ct].y + wrk_y;
				wrk_z = atm_vec[atm_ct].z + wrk_z;
		}	/* seq_ct */
	}	/* mod_ct */
	fprintf(STDOUT,
		"check_atom_position() : Atom position check done (atom_num= %d).\n",
		my_macro.atm_num);
	return SUCCESS;
}

struct pe_vector * set_atom_vectors(void)
{
	struct pe_vector * atm_vec;
	int mod_ct, seq_ct;
	
	atm_vec = (struct pe_vector *) malloc((my_macro.atm_num + 1)
					      * sizeof(struct pe_vector));
	if (atm_vec == NULL) {
		fprintf(STDERR, 
			"set_atom_vectors() : ERROR allocation failed\n");
		exit(1);
	}
	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for (seq_ct = my_macro.model[mod_ct].seq_head;
		     seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			if (compute_atom(seq_ct, atm_vec) == FAIL) exit(1);
		}
	}
	if (0) if(check_atom_position(atm_vec) == FAIL) exit(1);
	return atm_vec;
}

int compute_translation_vector(int seq_ct, struct pe_vector * t_vec, struct pe_slt_tile * s_tiles)
{
	int t_ct;
	float wrk_x, wrk_y, wrk_z;
	
	for (t_ct = my_macro.sequence[seq_ct].tile_head;
	     t_ct < my_macro.sequence[seq_ct].tile_tail; t_ct++) {
		wrk_x = s_tiles[t_ct + 1].cx - s_tiles[t_ct].cx;
		wrk_y = s_tiles[t_ct + 1].cy - s_tiles[t_ct].cy;
		wrk_z = s_tiles[t_ct + 1].cz - s_tiles[t_ct].cz;
		t_vec[t_ct].x = wrk_x;
		t_vec[t_ct].y = wrk_y;
		t_vec[t_ct].z = wrk_z;
		t_vec[t_ct].length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
					  + wrk_z * wrk_z);
		if (!FLOAT_NEQ(t_vec[t_ct].length)) {
			fprintf(STDERR,
				"compute_translation_vector() : ERROR t_vec[%d] = s_tile[%d].c - s_tile[%d].c == (0,0,0),\n",
				t_ct, t_ct + 1, t_ct);
			fprintf(STDERR,
				"compute_translation_vector() :       where s_tile[%d].c = (%f, %f, %f)\n",
				t_ct+1, s_tiles[t_ct + 1].cx,
				s_tiles[t_ct + 1].cy, s_tiles[t_ct + 1].cz);
			fprintf(STDERR,
				"compute_translation_vector() :             s_tile[%d].c = (%f, %f, %f)\n",
				t_ct, s_tiles[t_ct].cx, s_tiles[t_ct].cy,
				s_tiles[t_ct].cz);
			return FAIL;
		}
		if (0) {
			fprintf(STDERR,
				"compute_translation_vector() : DEBUG s_tile[%d].c = (%f, %f, %f)\n",
				t_ct+1, s_tiles[t_ct + 1].cx,
				s_tiles[t_ct + 1].cy, s_tiles[t_ct + 1].cz);
		}
	}
	if (t_ct < my_macro.tile_num) {
		wrk_x = s_tiles[t_ct + 1].cx - s_tiles[t_ct].cx;
		wrk_y = s_tiles[t_ct + 1].cy - s_tiles[t_ct].cy;
		wrk_z = s_tiles[t_ct + 1].cz - s_tiles[t_ct].cz;
		t_vec[t_ct].x = wrk_x;
		t_vec[t_ct].y = wrk_y;
		t_vec[t_ct].z = wrk_z;
		t_vec[t_ct].length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
					  + wrk_z * wrk_z);
		if (!FLOAT_NEQ(t_vec[t_ct].length)) {
			fprintf(STDERR,
				"compute_translation_vector() : ERROR t_vec[%d] = s_tile[%d].c - s_tile[%d].c == (0,0,0),\n",
				t_ct, t_ct + 1, t_ct);
			fprintf(STDERR,
				"compute_translation_vector() :       where s_tile[%d].c = (%f, %f, %f)\n",
				t_ct+1, s_tiles[t_ct + 1].cx,
				s_tiles[t_ct + 1].cy, s_tiles[t_ct + 1].cz);
			fprintf(STDERR,
				"compute_translation_vector() :             s_tile[%d].c = (%f, %f, %f)\n",
				t_ct, s_tiles[t_ct].cx, s_tiles[t_ct].cy,
				s_tiles[t_ct].cz);
			return FAIL;
		}
	} else {
		t_vec[t_ct].x = 0.0;
		t_vec[t_ct].y = 0.0;
		t_vec[t_ct].z = 0.0;
		t_vec[t_ct].length = 0.0;
	}
	return SUCCESS;
}

/* t_vec[t_ct] = s_tiles[t_ct + 1] - s_tiles[t_ct] */
int check_tile_center(int seq_ct, struct pe_vector * t_vec,
		 struct pe_slt_tile * s_tiles)
{
	int t_ct;
	float wrk_x, wrk_y, wrk_z;
	
	for (t_ct = my_macro.sequence[seq_ct].tile_head;
	     t_ct < my_macro.sequence[seq_ct].tile_tail; t_ct++) {
		if (s_tiles[t_ct].code == C_UP) {
			wrk_x = my_trn_table[s_tiles[t_ct].type].upward_cx
				- my_trn_table[s_tiles[t_ct].type].cx;
			wrk_y = my_trn_table[s_tiles[t_ct].type].upward_cy
				- my_trn_table[s_tiles[t_ct].type].cy;
			wrk_z = my_trn_table[s_tiles[t_ct].type].upward_cz
				- my_trn_table[s_tiles[t_ct].type].cz;
		} else {
			wrk_x = my_trn_table[s_tiles[t_ct].type].downward_cx
				- my_trn_table[s_tiles[t_ct].type].cx;
			wrk_y = my_trn_table[s_tiles[t_ct].type].downward_cy
				- my_trn_table[s_tiles[t_ct].type].cy;
			wrk_z = my_trn_table[s_tiles[t_ct].type].downward_cz
				- my_trn_table[s_tiles[t_ct].type].cz;
		}
		if (FLOAT_NEQ(t_vec[t_ct].x - wrk_x)
		    || FLOAT_NEQ(t_vec[t_ct].y - wrk_y)
		    || FLOAT_NEQ(t_vec[t_ct].z - wrk_z) ) {
			fprintf(STDERR, 
				"check_tile_center() : Error t_vec[%d] = (%8.3f, %8.3f, %8.3f) != (%8.3f, %8.3f, %8.3f)\n",
				t_ct, t_vec[t_ct].x, t_vec[t_ct].y,
				t_vec[t_ct].z, wrk_x, wrk_y, wrk_z);
			if (0) print_t_vec(t_vec);
			return FAIL;
		}
	}
	if (t_ct < my_macro.tile_num) {
		wrk_x = s_tiles[t_ct + 1].cx - s_tiles[t_ct].cx;
		wrk_y = s_tiles[t_ct + 1].cy - s_tiles[t_ct].cy;
		wrk_z = s_tiles[t_ct + 1].cz - s_tiles[t_ct].cz;
		if (FLOAT_NEQ(t_vec[t_ct].x - wrk_x)
		    || FLOAT_NEQ(t_vec[t_ct].y - wrk_y)
		    || FLOAT_NEQ(t_vec[t_ct].z - wrk_z) ) {
			fprintf(STDERR, 
				"check_tile_center() : Error t_vec[%d] = (%8.3f, %8.3f, %8.3f) != (%8.3f, %8.3f, %8.3f) [tail]\n",
				t_ct, t_vec[t_ct].x, t_vec[t_ct].y,
				t_vec[t_ct].z, wrk_x, wrk_y, wrk_z);
			return FAIL;
		}
	}
	return SUCCESS;
}

struct pe_vector * set_tile_vectors(struct pe_slt_tile * s_tiles)
{
	struct pe_vector * t_vec;
	int mod_ct, seq_ct;

	t_vec = (struct pe_vector *) malloc((my_macro.tile_num + 1)
					    * sizeof(struct pe_vector) );
	if (t_vec == NULL) {
		fprintf(STDERR, 
			"set_tile_vectors() : ERROR allocation failed\n");
		exit(1);
	}
	if (0) print_tiles(s_tiles);
	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for (seq_ct = my_macro.model[mod_ct].seq_head;
		     seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			if (compute_translation_vector(seq_ct, t_vec, s_tiles)
			    == FAIL)
				exit(1);
		}
	}

	if (DEBUG && my_macro.rotate_tile == PV_FALSE) {
		for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
			for (seq_ct = my_macro.model[mod_ct].seq_head;
			     seq_ct <= my_macro.model[mod_ct].seq_tail;
			     seq_ct++) {
				if (check_tile_center(seq_ct, t_vec, s_tiles)
				    == FAIL)
					exit(1);
			}
		}
		if (0) fprintf(STDERR, 
			"set_tile_vectors() : DEBUG center check done\n");
	}
	return t_vec;
}

/*
   Mtile_coor[16]
   (my_coor.axes[0] my_coor.axes[1] my_coor.axes[2]) <= M (X Y Z)
   where X= (1,1,1), Y= (-1,-1,1), Z= (1, -1, -1), W= (-1,1,-1)
*/
float set_mtile_coor(void)
{
	float factor;
	float detM;
	
	factor = sqrt(3.0)/my_coor.axes[0].length;	
	Mtile_coor[0] = (my_coor.axes[0].x/2.0 + my_coor.axes[2].x/2.0)*factor;
	Mtile_coor[1] = (my_coor.axes[0].y/2.0 + my_coor.axes[2].y/2.0)*factor;
	Mtile_coor[2] = (my_coor.axes[0].z/2.0 + my_coor.axes[2].z/2.0)*factor;
	Mtile_coor[3] = 0.0;
	Mtile_coor[4] =  (-my_coor.axes[1].x/2.0 - my_coor.axes[2].x/2.0)*factor;
	Mtile_coor[5] =  (-my_coor.axes[1].y/2.0 - my_coor.axes[2].y/2.0)*factor;
	Mtile_coor[6] =  (-my_coor.axes[1].z/2.0 - my_coor.axes[2].z/2.0)*factor;
	Mtile_coor[7] = 0.0;
	Mtile_coor[8] =  (my_coor.axes[0].x/2.0 + my_coor.axes[1].x/2.0)*factor;
	Mtile_coor[9] =  (my_coor.axes[0].y/2.0 + my_coor.axes[1].y/2.0)*factor;
	Mtile_coor[10] = (my_coor.axes[0].z/2.0 + my_coor.axes[1].z/2.0)*factor;
	Mtile_coor[11] = 0.0;
	Mtile_coor[12] = 0.0;
	Mtile_coor[13] = 0.0;
	Mtile_coor[14] = 0.0;
	Mtile_coor[15] = 1.0;
	/* det M(1,1)...M(3,3) */
	detM = Mtile_coor[0] * (Mtile_coor[5] * Mtile_coor[10]
				- Mtile_coor[9] * Mtile_coor[6])
		- Mtile_coor[4] * (Mtile_coor[1]*Mtile_coor[10]
				   - Mtile_coor[9] * Mtile_coor[2])
		+ Mtile_coor[8] * (Mtile_coor[1] * Mtile_coor[6]
				   - Mtile_coor[5] * Mtile_coor[2]);
	if (0) fprintf(STDERR, "set_mtile_coor():  detM= %8.3f\n", detM);
	return detM;
}

/*
   Tcenter[3]
   The center of a protein.
*/
void set_tcenter_and_ssize(void)
{
	int mod_ct, seq_ct, atm_ct;
	float max_x, min_x;
	float max_y, min_y;
	float max_z, min_z;
	float tmp;
	
	min_x = my_macro.atom[1].x;
	min_y = my_macro.atom[1].y;
	min_z = my_macro.atom[1].z;
	max_x = min_x;
	max_y = min_y;
	max_z = min_z;
	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for ( seq_ct= my_macro.model[mod_ct].seq_head;
		      seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			for (atm_ct= my_macro.sequence[seq_ct].atom_head;
			     atm_ct <= my_macro.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				if (max_x < my_macro.atom[atm_ct].x)
					max_x = my_macro.atom[atm_ct].x;
				if (min_x > my_macro.atom[atm_ct].x)
					min_x = my_macro.atom[atm_ct].x;
				if (max_y < my_macro.atom[atm_ct].y)
					max_y = my_macro.atom[atm_ct].y;
				if (min_y > my_macro.atom[atm_ct].y)
					min_y = my_macro.atom[atm_ct].y;
				if (max_z < my_macro.atom[atm_ct].z)
					max_z = my_macro.atom[atm_ct].z;
				if (min_z > my_macro.atom[atm_ct].z)
					min_z = my_macro.atom[atm_ct].z;
			}
		}
		break;	/* Consider only model one. */
	}
	Tcenter[0] = -1.0 * (max_x + min_x) / 2.0;
	Tcenter[1] = -1.0 * (max_y + min_y) / 2.0;
	Tcenter[2] = -1.0 * (max_z + min_z) / 2.0;
	Tcenter_save[0] = Tcenter[0];
	Tcenter_save[1] = Tcenter[1];
	Tcenter_save[2] = Tcenter[2];
	
	/* Ssize[4] */
	Ssize[0] = (max_x - min_x);
	Ssize[1] = (max_y - min_y);
	Ssize[2] = (max_z - min_z);
	if ( Ssize[0] >= Ssize[1])	tmp = Ssize[0];
	else				tmp = Ssize[1];
	if ( tmp >= Ssize[2])		Ssize[3] = tmp;
	else 				Ssize[3] = Ssize[2];
	if (0) {
		fprintf(STDOUT, "set_matrices():  x ¥in [%8.3f, %8.3f]\n",
			min_x, max_x);
		fprintf(STDOUT, "set_matrices():  y ¥in [%8.3f, %8.3f]\n",
			min_y, max_y);
		fprintf(STDOUT, "set_matrices():  z ¥in [%8.3f, %8.3f]\n",
			min_z, max_z);
		fprintf(STDOUT,
			"set_matrices():  Center = (%8.3f, %8.3f, %8.3f)\n",
			-1.0 * Tcenter[0], -1.0 * Tcenter[1], -1.0 *Tcenter[2]);
		fprintf(STDOUT,
			"set_matrices():  Size = %8.3f x %8.3f x %8.3f (Ssize[3]=%8.3f)\n",
			Ssize[0], Ssize[1], Ssize[2], Ssize[3]);
	}
	return;
}

float set_matrices()
{
	float detM;

	detM = set_mtile_coor();	
	if ((detM * detM - 1.0) > 2.0 / 100.0) {
		fprintf(STDERR,
			"set_matrices(): ERROR  detM= %8.3f (|detM| != 1, |axe|= %8.3f)\n",
			detM, my_coor.axes[0].length);
		exit(1);
	}
	
	set_tcenter_and_ssize();
	
	if (detM > 0.0)	return 1.0;
	else		return -1.0;
}

/* The position of the tiles of the 'seq_ct'-th sequence */
void position_of_init_tile(struct pe_slt_tile *s_tiles,
			   struct pe_coor_system *pt_coor)
{
	int t_ct, seq_ct, mod_ct;

	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for (seq_ct = my_macro.model[mod_ct].seq_head;
		     seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			for (t_ct = my_macro.sequence[seq_ct].tile_head;
			     t_ct < my_macro.sequence[seq_ct].tile_tail;
			     t_ct++) {
				s_tiles[t_ct].tx = pt_coor->origin.x
				    + s_tiles[t_ct].top[0] * pt_coor->axes[0].x
				    + s_tiles[t_ct].top[1] * pt_coor->axes[1].x
				    + s_tiles[t_ct].top[2] * pt_coor->axes[2].x
				    + s_tiles[t_ct].top[3] * pt_coor->axes[3].x;
				s_tiles[t_ct].ty = pt_coor->origin.y
				    + s_tiles[t_ct].top[0] * pt_coor->axes[0].y
				    + s_tiles[t_ct].top[1] * pt_coor->axes[1].y
				    + s_tiles[t_ct].top[2] * pt_coor->axes[2].y
				    + s_tiles[t_ct].top[3] * pt_coor->axes[3].y;
				s_tiles[t_ct].tz = pt_coor->origin.z
				    + s_tiles[t_ct].top[0] * pt_coor->axes[0].z
				    + s_tiles[t_ct].top[1] * pt_coor->axes[1].z
				    + s_tiles[t_ct].top[2] * pt_coor->axes[2].z
				    + s_tiles[t_ct].top[3] * pt_coor->axes[3].z;
				s_tiles[t_ct].cx
					= my_trn_table[s_tiles[t_ct].type].cx
					+ s_tiles[t_ct].tx;
				s_tiles[t_ct].cy
					= my_trn_table[s_tiles[t_ct].type].cy
					+ s_tiles[t_ct].ty;
				s_tiles[t_ct].cz
					= my_trn_table[s_tiles[t_ct].type].cz
					+ s_tiles[t_ct].tz;
				if (0) {
					fprintf(STDERR,
"position_of_init_tile(): s_tile[1].top = (%d, %d, %d, %d), type= %s\n",
					    s_tiles[1].top[0],
					    s_tiles[1].top[1],
					    s_tiles[1].top[2],
					    s_tiles[1].top[3],
					    my_trn_table[s_tiles[1].type].str);
					fprintf(STDERR,
"position_of_init_tile():               = (%8.3f, %8.3f, %8.3f)\n",
						s_tiles[1].tx,  s_tiles[1].ty,
						s_tiles[1].tz);
				}
			}
		}
	}	/* mod_ct */
	return;
}

void myinit(char ** argv, struct pe_slt_tile ** p_tiles,
	    struct pe_vector ** p_vectors, struct pe_vector **p_tvectors)
{
	float detM_flg;
	struct pe_atom *atoms;
	struct pe_slt_tile *stiles;
	struct pe_vector *vectors;
	struct pe_vector *tvectors;
	float atom_dis;
	float wrk_vx, wrk_vy, wrk_vz;

	/* Read data */
	init_macro();
	init_coor(&my_coor);
	stiles = read_data(argv, atoms, &my_coor);
	*p_tiles = stiles;

	/* Set uo the transition table: my_trn_table[24] */
	wrk_vx = atoms[2].x - atoms[1].x;
	wrk_vy = atoms[2].y - atoms[1].y;
	wrk_vz = atoms[2].z - atoms[1].z;
	atom_dis = sqrt(wrk_vx * wrk_vx + wrk_vy * wrk_vy + wrk_vz * wrk_vz);
	setup_transition_table(atom_dis);
	if (0) print_trn_table();

	if (0) print_macro(&my_macro);
	if (0) {
		print_atoms(atoms);
		print_coordinates(&my_coor);
		print_tiles(stiles);
	}
 
	/* Set uo the n-tile path table: npath_table[24] */
	my_settings.tiles_per_res = my_macro.tiles_per_res;
	my_settings.along_backbone = my_macro.along_backbone;
	my_settings.rotate_tile = my_macro.rotate_tile;
	my_settings.max_path = (int) pow(2.0, (double)
					 my_settings.tiles_per_res);
	if (0) fprintf(STDOUT, "myinit() :  DEBUG [TODO] my_settings\n");
	if(setup_npath_table() == FAIL) {
		print_npath_table();
		exit(1);
	}

	/* Set up atom vectors */
	vectors = set_atom_vectors();
	if (0) print_vectors(vectors);
	*p_vectors = vectors;

/*
	position_of_init_tile(stiles, &my_coor);
	if (DEBUG) {
		fprintf(STDERR,
			"myInit(): s_tile[1].top = (%d, %d, %d, %d), type= %s\n",
			stiles[1].top[0], stiles[1].top[1],
			stiles[1].top[2], stiles[1].top[3], my_trn_table[stiles[1].type].str);
		fprintf(STDERR,
"myInit(): s_tile[1].t*  = (%8.3f, %8.3f, %8.3f)\n", stiles[1].tx,  stiles[1].ty,  stiles[1].tz);
		fprintf(STDERR,
"myInit(): s_tile[1].c*  = (%8.3f, %8.3f, %8.3f)\n", stiles[1].cx,  stiles[1].cy,  stiles[1].cz);
	}
*/

	/* Set up tile vectors */
	tvectors = set_tile_vectors(stiles);
	if (0) print_vectors(tvectors);
	*p_tvectors = tvectors;

	/* Set Mtile_coor[], Tcenter[], and Ssize matrices */
	detM_flg = set_matrices();
	if (0) {
		fprintf(STDERR, "myInit(): ***Matrices***\n");
		fprintf(STDERR,
			"myInit():             |%8.3f %8.3f %8.3f %8.3f|\n",
			Mtile_coor[0], Mtile_coor[4], Mtile_coor[8],
			Mtile_coor[12]);
		fprintf(STDERR,
			"myInit(): Mtile_coor= |%8.3f %8.3f %8.3f %8.3f|\n",
			Mtile_coor[1], Mtile_coor[5], Mtile_coor[9],
			Mtile_coor[13]);
		fprintf(STDERR,
			"myInit():             |%8.3f %8.3f %8.3f %8.3f|\n",
			Mtile_coor[2], Mtile_coor[6], Mtile_coor[10],
			Mtile_coor[14]);
		fprintf(STDERR,
			"myInit():             |%8.3f %8.3f %8.3f %8.3f|\n",
			Mtile_coor[3], Mtile_coor[7], Mtile_coor[11],
			Mtile_coor[15]);
		fprintf(STDERR,
			"myInit():             (detM_flg = %8.3f)\n", detM_flg);
		fprintf(STDERR,
			"myInit(): Tcenter=    (%8.3f, %8.3f, %8.3f)\n",
			Tcenter[0], Tcenter[1], Tcenter[2]);
		fprintf(STDERR,
			"myInit(): Ssize=      (%8.3f, %8.3f, %8.3f)\n",
			Ssize[0], Ssize[1], Ssize[2]);
		fprintf(STDERR, "myInit(): ***\n");
	}
	
	if (my_settings.rotate_tile == PV_TRUE)
		set_rotation_matrix(stiles);
	create_display_list();

	if (0) glClearColor (0.7, 0.7, 0.85, 1.0);	/* pale blue */
	else if (0) glClearColor (0.7, 0.7, 0.7, 1.0);	/* gray */
	else glClearColor (1.0, 1.0, 1.0, 1.0);		/* white */
	glClearDepth(1.0);

	glLightfv(GL_LIGHT0,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT0,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT1,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT1,GL_AMBIENT,light_ambient);
	glLightfv(GL_LIGHT2,GL_DIFFUSE,light_diffuse);
	glLightfv(GL_LIGHT2,GL_AMBIENT,light_ambient);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT,lmodel_ambient);
	if(1) {
		light_position[3] = 0.0F;	/* directional source */
		light_position2[3] = 0.0F;	/* directional source */
		light_position3[3] = 0.0F;	/* directional source */
	} else {
    		light_position[3] = 1.0F;
		light_position2[3] = 1.0F;
		light_position3[3] = 1.0F;
	}
	glLightfv(GL_LIGHT0,GL_POSITION,light_position);
	glLightfv(GL_LIGHT1,GL_POSITION,light_position2);
	glLightfv(GL_LIGHT2,GL_POSITION,light_position3);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	glEnable(GL_LIGHT2);
	glEnable(GL_CULL_FACE);
	glEnable(GL_NORMALIZE);

	glEnable(GL_BLEND);
	if (0) glDisable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	if (0) glEnable(GL_DEPTH_TEST);

	if (SHOW_COOR_AXES)
		fprintf(STDOUT,
			"myInit(): [NOTE] Coordinate axes indicater <x-axis=green ball, y-axis=blue ball, z-axis=red ball>.\n");
	if (0) fprintf(STDOUT,
		"myInit(): [NOTE] Model-one only is rendered.\n");
	fprintf(STDOUT,
		"myInit(): [NOTE] Press 'B' to enable/disable Blending.\n");
	fprintf(STDOUT,
		"myInit(): [NOTE] Press 'S' to render CYLINDER/SOLID TETRA./WIRE TETRA.\n");
	fprintf(STDOUT,
		"myInit(): [NOTE] Press 'M' to render BOTH_PRT_AND_MDL/PROTEIN_ONLY/MODEL_ONLY.\n");
	fprintf(STDOUT,
		"myInit(): [NOTE] Press 'Z' to zoom in and 'z' to zoom out.\n");
	fprintf(STDOUT,
		"myInit(): [NOTE] Press 'T' to translate (controlled by mouse move).\n");
	fprintf(STDOUT,
		"myInit(): [NOTE] Press 'R' to rotate (controlled by mouse move).\n");
	fprintf(STDOUT,
		"myInit(): [NOTE] Press 'C' to undo all.\n");
	return;
}
