/*
 *  pv_read_data.h
 *  ProteinViewer
 *
 *  Created by Naoto Morikawa on 6/17/06.
 *  Copyright 2006 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 */

/*******************************************/
/*    Chain A    Molecule 1    Model 1     */
/*    --------                             */
/*    Chain B                              */
/*    ---------------------                */
/*    Chain C    Molecule 2                */
/*    --------                             */
/*    Chain D                              */
/*    ----------------------------------   */
/*    Chain A    Molecule 1    Model 2     */
/*    --------                             */
/*    Chain B                              */
/*    ---------------------                */
/*    Chain C    Molecule 2                */
/*    --------                             */
/*    Chain D                              */
/*******************************************/

struct pe_model {
	int seq_head;
	int seq_tail;
};

struct pe_molecule {
	int chain_head;
	int chain_tail;
	int id;      /* COMPND (1, 2, ... ) */
	char  name[LineWidth+1];
};

struct pe_chain {
	int atom_head;
	int atom_tail;
	char id;        /* COMPND (A, B, ...) */
	int mol_id;
	int mod_num;
	int tile_head;
	int tile_tail;
};

struct pe_atom {
	int serial_num;          /* column 7-11 */
	PE_ATOM_NAME atom_name;            /* column 13-16: CA, C, N, ... */
	char alt_loc;             /* column 17 */
	char res_name[3];         /* column 18-20 */
	char chain_id;            /* column 22 */
	int res_seq;             /* column 23-26 */
	float x;                  /* coloum 31-38  Real(8.3)*/
	float y;                  /* coloum 39-46 */
	float z;                  /* coloum 47-54 */
	float occupancy;          /* coloum 55-60  Real(6.2)*/
	char seg_id[4];           /* coloum 73-76 */
	float err;
	char flag;
};

#define MOL_SIZE 100
#define CHN_SIZE 100

struct pv_macro {
	int mod_num;
	struct pe_model *model;
	int mol_num;             /* up to 9 (1, 2, ..., 9) */
	struct pe_molecule molecule[100];   /* 1 + 99 */
	int chn_num;             /* up to 24 (A, B, ..., Z) */
	struct pe_chain chain[100];         /*  1 + 98 + 1 */
	int res_num;
	int atm_num;
	struct pe_atom *atom;
	struct pe_atom *atom_saved;
	int seq_num;
	struct pe_chain *sequence;
	int tile_max_num;
	int tile_num;
	float dist_factor;	/* |pos_a - pos_b| < dist_factor*ave_dis */
	int error_tab_size;
	float tiles_per_res;        /* #tile/#residues */
	int along_backbone;
	PV_BOOLEAN rotate_tile;
};
