/*
 *  pv_read_data_x11.c (for X11)
 *  ProteinViewer
 *
 *  Created by Naoto Morikawa on 6/17/06.
 *  Copyright 2006 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-03 support the case of "my_macro.tile_num == 0"
 *  2006-11-20 version 0.2 release
 */

#include <stdlib.h>
#include <GL/glut.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "pv_main.h"
#include "pv_read_data.h"

extern struct pv_macro my_macro;

void print_molecule(struct pe_molecule *pt_strct)
{
	fprintf(STDOUT, "mol. : [chn_head, chn_tail] id name\n");
	fprintf(STDOUT, "mol. : [ %7d,  %7d]  %c %s\n", pt_strct->chain_head,
		pt_strct->chain_tail, pt_strct->id, pt_strct->name);
	return;
}

void print_chain(struct pe_chain *pt_strct)
{
	fprintf(STDOUT,
		"chain: [atm_head, atm_tail] chn_id |mol_id |mod_num [tile_head,tile_tail]\n");
	fprintf(STDOUT, "chain: [ %7d,  %7d] %7c|%7d|%7d [ %8d, %8d]\n",
		pt_strct->atom_head, pt_strct->atom_tail, pt_strct->id,
		pt_strct->mol_id, pt_strct->mod_num, pt_strct->tile_head,
		pt_strct->tile_tail);
	return;
}

int tiletype_to_int(char *tile_type)
{
	switch(*tile_type) {
		case('x'):
			switch(*(tile_type+1) ) {
				case('y'):
					switch(*(tile_type+2) ) {
						case('z'): return XYZW;
						case('w'): return XYWZ;
						default: return -1;
					}
				case('z'):
					switch(*(tile_type+2) ) {
						case('y'): return XZYW;
						case('w'): return XZWY;
						default: return -1;
					}
				case('w'):
					switch(*(tile_type+2) ) {
						case('y'): return XWYZ;
						case('z'): return XWZY;
						default: return -1;
					}
				default:
					return -1;
			}
		case('y'):
			switch(*(tile_type+1) ) {
				case('x'):
					switch(*(tile_type+2) ) {
						case('z'): return YXZW;
						case('w'): return YXWZ;
						default: return -1;
					}
				case('z'):
					switch(*(tile_type+2) ) {
						case('x'): return YZXW;
						case('w'): return YZWX;
						default: return -1;
					}
				case('w'):
					switch(*(tile_type+2) ) {
						case('x'): return YWXZ;
						case('z'): return YWZX;
						default: return -1;
					}
				default:
					return -1;
			}
		case('z'):
			switch(*(tile_type+1) ) {
				case('x'):
					switch(*(tile_type+2) ) {
						case('y'): return ZXYW;
						case('w'): return ZXWY;
						default: return -1;
					}
				case('y'):
					switch(*(tile_type+2) ) {
						case('x'): return ZYXW;
						case('w'): return ZYWX;
						default: return -1;
					}
				case('w'):	switch(*(tile_type+2) ) {
					case('x'): return ZWXY;
					case('y'): return ZWYX;
					default: return -1;
				}
				default:
					return -1;
			}
		case('w'):
			switch(*(tile_type+1) ) {
				case('y'):
					switch(*(tile_type+2) ) {
						case('x'): return WYXZ;
						case('z'): return WYZX;
						default: return -1;
					}
				case('z'):
					switch(*(tile_type+2) ) {
						case('x'): return WZXY;
						case('y'): return WZYX;
						default: return -1;
					}
				case('x'):
					switch(*(tile_type+2) ) {
						case('y'): return WXYZ;
						case('z'): return WXZY;
						default: return -1;
					}
				default:
					return -1;
			}
		default:
			return -1;
	}
}

struct pe_atom * read_pdb_data_line(char * buf)
{
	int i;
	struct pe_atom * atoms;
	
	if(!strncmp("<PDB_data ", buf, 10)) {
		if ( sscanf(buf,
"<PDB_data atoms=\"%d\" residues=\"%d\" sequences=\"%d\" molecules=\"%d\" models=\"%d\">\n",
			    &(my_macro.atm_num), &(my_macro.res_num),
			    &(my_macro.seq_num), &(my_macro.mol_num),
			    &(my_macro.mod_num) ) == EOF) {
			fprintf(STDERR, "read_data(): Format error in %s\n",
				buf);
			exit(1);
		}
		if (DEBUG) fprintf(STDOUT,
"read_data(): atm_num= %d, res_num= %d, seq_num= %d, mol_num= %d, mod_num= %d\n",
				   my_macro.atm_num,
				   my_macro.res_num,
				   my_macro.seq_num,
				   my_macro.mol_num,
				   my_macro.mod_num);
		/* Add 'my_macro.seq_num' for the case of empty seq. */
		atoms = (struct pe_atom *) malloc((my_macro.atm_num + 1
						   + my_macro.seq_num)
						  * sizeof(struct pe_atom));
		if (atoms == NULL) {
			fprintf(STDERR, 
				"read_pdb_data_line() : ERROR atoms[] allocation failed\n");
			exit(1);
		}
		/* Initialization */
		for (i = 0; i <= my_macro.atm_num; i++) {
			atoms[i].serial_num = 0;
			atoms[i].atom_name = N_UNKNOWN;	/* Mr. Nobody! */
			strncpy(atoms[i].res_name, "xxx", 3);
			atoms[i].chain_id = 'x';
			atoms[i].res_seq = 0;
			atoms[i].x = 0.0;
			atoms[i].y = 0.0;
			atoms[i].z = 0.0;
			atoms[i].occupancy = 0.0;
			strncpy(atoms[i].seg_id, "xxxx", 4);
			/* atoms[i].flag = 0; */	/* Not used */
		}
		my_macro.atom = atoms;
		
		/* 0-th + model + dummy */
		my_macro.model = (struct pe_model *)
			malloc((my_macro.mod_num + 2)*sizeof(struct pe_model));
		if (my_macro.model == NULL) {
			fprintf(STDERR, 
				"read_pdb_data_line() : ERROR model[] allocation failed\n");
			exit(1);
		}
		for (i = 0; i < my_macro.mod_num + 2; i++) {
			my_macro.model[i].seq_head = 0;
			my_macro.model[i].seq_tail = 0;
		}
		my_macro.model[1].seq_head = 1;
		
		/* 0-th + sequence + dummy */
		my_macro.sequence = (struct pe_chain *)
			malloc((my_macro.seq_num +2) * sizeof(struct pe_chain));
		if (my_macro.sequence == NULL) {
			fprintf(STDERR, 
				"read_pdb_data_line() : ERROR sequence[] allocation failed\n");
			exit(1);
		}
		for (i = 0; i < my_macro.seq_num + 2; i++) {
			my_macro.sequence[i].atom_head = 0;
			my_macro.sequence[i].atom_tail = 0;
			my_macro.sequence[i].id = '-';
			my_macro.sequence[i].mol_id = 0;
			my_macro.sequence[i].mod_num = 0;
			my_macro.sequence[i].tile_head = 0;
			my_macro.sequence[i].tile_tail = 0;
		}
	}
	return atoms;
}

struct pe_slt_tile * read_slant_tiles_line(char * buf)
{
	int i;
	struct pe_slt_tile * s_tiles;
	
	if(!strncmp("<Slant_tiles ", buf, 13)) {
		if ( sscanf(buf,"<Slant_tiles tiles=\"%d\">\n",
			    &(my_macro.tile_num) ) == EOF) {
			fprintf(STDERR, "read_data(): Format error in %s\n",
				buf);
			exit(1);
		}
		if (DEBUG) fprintf(STDOUT, "read_data(): tile_num= %d\n",
				   my_macro.tile_num);
		
		/* Add 'my_macro.seq_num' for the case of empty seq. */
		s_tiles = (struct pe_slt_tile *) malloc((my_macro.tile_num + 1
			+ my_macro.seq_num) * sizeof(struct pe_slt_tile) );
		if (s_tiles == NULL) {
			fprintf(STDERR, 
				"read_slant_tiles_line() : ERROR allocation failed\n");
			exit(1);
		}
		/* Initialization */
		for (i = 0; i <= my_macro.tile_num; i++) {
			s_tiles[i].cx = 0.0;
			s_tiles[i].cy = 0.0;
			s_tiles[i].cz = 0.0;
			s_tiles[i].tx = 0.0;
			s_tiles[i].ty = 0.0;
			s_tiles[i].tz = 0.0;
			s_tiles[i].error = -1.0;
			s_tiles[i].error_atom = 0;
			s_tiles[i].atom_flag = 0;
			s_tiles[i].top[0] = 0;
			s_tiles[i].top[1] = 0;
			s_tiles[i].top[2] = 0;
			s_tiles[i].top[3] = 0;
			s_tiles[i].type = -1;
			s_tiles[i].code = C_UNKNOWN;
			s_tiles[i].ter = 0;
		}
	}
	return s_tiles;
}

int read_tile_sequence_line(char * buf, int tile_seq_ct)
{
	int t_seq_ct;
	
	t_seq_ct = tile_seq_ct;
	if(!strncmp("<Tile_sequence ", buf, 15)) {
		t_seq_ct++;
	}
	return t_seq_ct;
}

int read_atom_sequence_line(char * buf, int sequence_ct, int a_ct)
{
	int seq_ct;
	
	seq_ct = sequence_ct;
	if(!strncmp("<Atom_sequence ", buf, 15)) {
		seq_ct++;
		if (seq_ct <= my_macro.seq_num)
			my_macro.sequence[seq_ct].atom_head = a_ct + 1;
	}
	return seq_ct;
}

/* atm_type <=> atom[a_ct].atom_name */
int read_atom_line(char * buf, struct pe_atom * atoms, int atom_ct, int seq_ct)
{
	int a_ct;
	int atm_ct;
	char atm_type[2+1];
	
	a_ct = atom_ct;
	if(!strncmp("<atom>", buf, 6)) {
		a_ct++;
		if (sscanf(buf,
			   "<atom> [%6d] %6d  %2s  %3s  %c  %4d (%8f, %8f, %8f) %4f %4s </atom>\n",
			   &atm_ct, &(atoms[a_ct].serial_num), atm_type,
			   atoms[a_ct].res_name, &(atoms[a_ct].chain_id),
			   &(atoms[a_ct].res_seq), &(atoms[a_ct].x),
			   &(atoms[a_ct].y), &(atoms[a_ct].z),
			   &(atoms[a_ct].occupancy),
			   atoms[a_ct].seg_id) == EOF) {
			fprintf(STDERR, "read_data(): Format error in %s\n",
				buf);
			exit(1);
		}
		
		if (!strncmp(atoms[a_ct].seg_id, "xxxx", 4)) {
			fprintf(STDERR,
				"read_data(): Read error at atom [%d]\n",
				a_ct);
			fprintf(STDOUT,
"<atom> [%6d] %6d  %2s  %.3s  %c  %4d (%8.3f, %8.3f, %8.3f) %4.3f %.4s </atom>\n",
				atm_ct, atoms[a_ct].serial_num, atm_type,
				atoms[a_ct].res_name, atoms[a_ct].chain_id,
				atoms[a_ct].res_seq, atoms[a_ct].x,
				atoms[a_ct].y, atoms[a_ct].z,
				atoms[a_ct].occupancy, atoms[a_ct].seg_id);
			exit(1);
		}
		if (atm_ct <= 0) {
			a_ct--;
			my_macro.sequence[seq_ct].atom_head = 0;
			return a_ct;
		}

		if(!strncmp("CA", atm_type, 2)) {
			atoms[a_ct].atom_name = N_CA;
		} else if(!strncmp("CR", atm_type, 2)) {
			atoms[a_ct].atom_name = N_CR;
		} else if(!strncmp("CY", atm_type, 2)) {
			atoms[a_ct].atom_name = N_CY;
		} else if(!strncmp("CW", atm_type, 2)) {
			atoms[a_ct].atom_name = N_CW;
		} else if(!strncmp("C", atm_type, 1)) {
			atoms[a_ct].atom_name = N_CC;
		} else if(!strncmp("N", atm_type, 1)) {
			atoms[a_ct].atom_name = N_NN;
		} else {
			atoms[a_ct].atom_name = N_UNKNOWN;
			fprintf(STDERR,
				"read_data(): Format error atm_type= %s\n",
				atm_type);
			exit(1);
		}
	}
	return a_ct;
}

int read_end_line(char * buf, int model_ct, int a_ct, int seq_ct, int t_ct,
		  struct pe_slt_tile * s_tiles)
{
	int mod_ct;
	
	mod_ct = model_ct;
	if(!strncmp("</PDB_data>", buf, 11)) {
		/* End of PDB data */
		mod_ct--;
	} else if(!strncmp("</Coordinates>", buf, 14)) {
		/* End of Coordinates */
	} else if(!strncmp("</Slant_tiles>", buf, 14)) {
		/* End of Slant tiles */
	} else if(!strncmp("</Atom_sequence>", buf, 16)) {
		/* End of atom sequence */
		if (seq_ct > 0)
			my_macro.sequence[seq_ct].atom_tail = a_ct;
	} else if(!strncmp("</model>",buf,8)) {
		/* End of model */
		my_macro.model[mod_ct].seq_tail = seq_ct;
		mod_ct++;
		if (mod_ct <= my_macro.mod_num)
			my_macro.model[mod_ct].seq_head = seq_ct + 1;
	} else if(!strncmp("</Tile_sequence>", buf, 16)) {
		/* End of tile sequence */
		s_tiles[t_ct].ter = 1;
	}
	return mod_ct;
}

void read_origine_line(char * buf, struct pe_coor_system * pt_coor)
{
	if (sscanf(buf, "<origin> ( %8f, %8f, %8f ) </origin>\n",
		   &(pt_coor->origin.x), &(pt_coor->origin.y),
		   &(pt_coor->origin.z)) == EOF) {
		fprintf(STDERR, "read_data(): Format error in %s\n", buf);
		exit(1);
	}
	if (0) fprintf(STDOUT, "<origin> ( %8.3f, %8.3f, %8.3f ) </origin>\n",
		       pt_coor->origin.x, pt_coor->origin.y,
		       pt_coor->origin.z);
	return;
}

void read_x_axis_line(char * buf, struct pe_coor_system * pt_coor)
{
	if (sscanf(buf,"<x-axis> ( %8f, %8f, %8f ) </x-axis>\n",
		   &(pt_coor->axes[0].x), &(pt_coor->axes[0].y),
		   &(pt_coor->axes[0].z)) == EOF) {
		fprintf(STDERR, "read_data(): Format error in %s\n", buf);
		exit(1);
	}
	pt_coor->axes[0].length = sqrt (pt_coor->axes[0].x * pt_coor->axes[0].x
				+ pt_coor->axes[0].y * pt_coor->axes[0].y
				+ pt_coor->axes[0].z * pt_coor->axes[0].z);
	if (0) fprintf(STDOUT, "<x-axis> ( %8.3f, %8.3f, %8.3f ) </x-axis>\n",
		       pt_coor->axes[0].x, pt_coor->axes[0].y,
		       pt_coor->axes[0].z);
	return;
}

void read_y_axis_line(char * buf, struct pe_coor_system * pt_coor)
{
	if (sscanf(buf,"<y-axis> ( %8f, %8f, %8f ) </y-axis>\n",
		   &(pt_coor->axes[1].x), &(pt_coor->axes[1].y),
		   &(pt_coor->axes[1].z)) == EOF) {
		fprintf(STDERR, "read_data(): Format error in %s\n", buf);
		exit(1);
	}
	pt_coor->axes[1].length = sqrt(pt_coor->axes[1].x * pt_coor->axes[1].x
				+ pt_coor->axes[1].y * pt_coor->axes[1].y
				+ pt_coor->axes[1].z * pt_coor->axes[1].z);
	return;
}

void read_z_axis_line(char * buf, struct pe_coor_system * pt_coor)
{
	if (sscanf(buf,"<z-axis> ( %8f, %8f, %8f ) </z-axis>\n",
		   &(pt_coor->axes[2].x), &(pt_coor->axes[2].y),
		   &(pt_coor->axes[2].z)) == EOF) {
		fprintf(STDERR, "read_data(): Format error in %s\n", buf);
		exit(1);
	}
	pt_coor->axes[2].length = sqrt(pt_coor->axes[2].x * pt_coor->axes[2].x
				+ pt_coor->axes[2].y * pt_coor->axes[2].y
				+ pt_coor->axes[2].z * pt_coor->axes[2].z);
	return;
}

void read_w_axis_line(char * buf, struct pe_coor_system * pt_coor)
{
	if (sscanf(buf,"<w-axis> ( %8f, %8f, %8f ) </w-axis>\n",
		   &(pt_coor->axes[3].x), &(pt_coor->axes[3].y),
		   &(pt_coor->axes[3].z)) == EOF) {
		fprintf(STDERR, "read_data(): Format error in %s\n", buf);
		exit(1);
	}
	pt_coor->axes[3].length = sqrt(pt_coor->axes[3].x * pt_coor->axes[3].x
				+ pt_coor->axes[3].y * pt_coor->axes[3].y
				+ pt_coor->axes[3].z * pt_coor->axes[3].z);
	return;
}

/* 
   tile_type <=> &(trn_table[s_tiles[t_ct].type].str)
   tile_code <=> s_tiles[i].code
   atm_type <=> my_macro.atom[s_tiles[i].error_atom].atom_name
 */
int read_tile_line(char * buf, struct pe_slt_tile * s_tiles, int init_t_ct)
{
	int t_ct;
	int tile_ct;
	char tile_type[3+1];
	char tile_code;
	char atm_type[2+1];
	
	t_ct = init_t_ct;
	if(!strncmp("<tile>", buf, 6)) {
		t_ct++;
		if (t_ct > my_macro.tile_num) return t_ct;
		if (sscanf(buf,
			   "<tile> [%d] x^%d y^%d z^%d w^%d [%3s], code= %c, atom= %d (%2s)\n",
			   &tile_ct, s_tiles[t_ct].top, s_tiles[t_ct].top + 1,
			   s_tiles[t_ct].top + 2, s_tiles[t_ct].top + 3,
			   tile_type, &tile_code, &(s_tiles[t_ct].error_atom),
			   atm_type) == EOF) {
			fprintf(STDERR, "read_data(): Format error in %s\n",
				buf);
			fprintf(STDOUT,
				"<tile> [%d] x^%d y^%d z^%d w^%d [%.3s], code= %c, atom= %d (%.2s)\n",
				tile_ct, s_tiles[t_ct].top[0],
				s_tiles[t_ct].top[1], s_tiles[t_ct].top[2], 
				s_tiles[t_ct].top[3], tile_type, tile_code,
				s_tiles[t_ct].error_atom, atm_type);
			exit(1);
		}
		if (tile_ct == 0) {
			t_ct--;
			return t_ct;
		}
		if( tile_code == 'U' ) {
			s_tiles[t_ct].code = C_UP;	/* UP = 1 */
		} else if( tile_code == 'D' ) {
			s_tiles[t_ct].code = C_DOWN;	/* DOWN = -1 */
		} else {
			fprintf(STDERR,
				"read_data(): Format error (tile_code != U or D) in %s\n",
				buf);
			exit(1);
		}
		s_tiles[t_ct].type = tiletype_to_int(tile_type);
		if (s_tiles[t_ct].type < 0) {
			fprintf(STDERR,
				"read_data(): Format error (type<0) in %s\n",
				buf);
			exit(1);
		}
		if (0) fprintf(STDOUT,
			       "<tile> [%d] x^%d y^%d z^%d w^%d [%.3s] (%d), code= %d (%c), atom= %d (%.2s)\n",
			       tile_ct, s_tiles[t_ct].top[0],
			       s_tiles[t_ct].top[1], s_tiles[t_ct].top[2], 
			       s_tiles[t_ct].top[3], tile_type,
			       s_tiles[t_ct].type, s_tiles[t_ct].code,
			       tile_code, s_tiles[t_ct].error_atom, atm_type);
	}
	return t_ct;
}

void read_space_line(char * buf, struct pe_slt_tile * s_tiles, int t_ct,
		     struct pe_coor_system * pt_coor)
{
	int error_atm;
	char code_type[7+1];
	
	if(!strncmp("       cen", buf, 10)) {
		if (t_ct > my_macro.tile_num) return;
		if (sscanf(buf,
			   "       center= (%8f, %8f, %8f), error= %f against atom [%d]\n",
			   &(s_tiles[t_ct].cx), &(s_tiles[t_ct].cy),
			   &(s_tiles[t_ct].cz), &(s_tiles[t_ct].error),
			   &error_atm) == EOF) {
			fprintf(STDERR, "read_data(): Format error in %s\n",
				buf);
			exit(1);
		}
	} else if(!strncmp("       dir", buf, 10)) {
		if (t_ct > my_macro.tile_num) return;
		if (sscanf(buf,
			   "       direction= (%8f, %8f, %8f) </tile>\n",
			   &(s_tiles[t_ct].dx), &(s_tiles[t_ct].dy),
			   &(s_tiles[t_ct].dz)) == EOF) {
			fprintf(STDERR, "read_data(): Format error in %s\n",
				buf);
			exit(1);
		}
		if (my_macro.rotate_tile == PV_FALSE)
			my_macro.rotate_tile = PV_TRUE;
	} else if(!strncmp("  Num. of ", buf, 10)) {
		if (sscanf(buf,
			   "  Num. of tiles per residue = %4f    (%7c\n",
			   &(my_macro.tiles_per_res), code_type) == EOF) {
			fprintf(STDERR, "read_data(): Format error in %s\n",
				buf);
			exit(1);
		}
		if (code_type[0] == '1') {
			my_macro.along_backbone = 0;
		} else if (code_type[6] == ')') {
			my_macro.along_backbone = 0;
		} else {
			my_macro.along_backbone = 1;
		}
	} else if(!strncmp("  Average d", buf, 11)) {
		if (sscanf(buf,
			   "  Average distance between atoms = %8f\n",
			   &(pt_coor->ave_dis)) == EOF) {
			fprintf(STDERR, "read_data(): Format error in %s\n",
				buf);
			exit(1);
		} else if (sscanf(buf,
				  "  Average distance between CA atoms = %8f\n",
				  &(pt_coor->ave_dis)) == EOF) {
			fprintf(STDERR, "read_data(): Format error in %s\n",
				buf);
			exit(1);
		}
	}
	return;
}

/* Set tile_head and tile_tail for each chain */
int set_tile_head_and_tail(struct pe_slt_tile * s_tiles)
{
	int t_ct, mod_ct, seq_ct;
	
	t_ct = 1;
	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for ( seq_ct = my_macro.model[mod_ct].seq_head;
		      seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			if (my_macro.sequence[seq_ct].atom_head <= 0) {
				/* DNA or RNA */
				my_macro.sequence[seq_ct].atom_tail = 0;
				my_macro.sequence[seq_ct].tile_head = 0;
				my_macro.sequence[seq_ct].tile_tail = 0;
			} else {
				if (my_macro.tile_num <= 0) continue;	/* 2006-11-03 */
				my_macro.sequence[seq_ct].tile_head = t_ct;
				while ( (t_ct < my_macro.tile_num)
					&& (s_tiles[t_ct].ter == 0) ) t_ct++;
				if (t_ct > my_macro.tile_num ) {
					fprintf(STDERR,
						"set_tile_head_and_tail() : my_macro.sequence[%d].tile_tail overflow (t_ct = %d, my_macro.tile_num = %d)\n",
						seq_ct, t_ct, my_macro.tile_num);	/* 2006-11-03 */
					return FAIL;
				}
				my_macro.sequence[seq_ct].tile_tail = t_ct;
				t_ct++;
			}
		}
	}
	return SUCCESS;
}

/* Read the pcd file */
struct pe_slt_tile * read_data(char ** argv, struct pe_atom * atoms,
			       struct pe_coor_system * pt_coor)
{
	char buf[LineWidth];
	FILE * fp_pcd;
	int mod_ct, seq_ct, t_seq_ct;
	int a_ct, t_ct;
	struct pe_slt_tile * s_tiles;
	
	if( (fp_pcd=fopen(argv[1],"r")) == NULL){    
		fprintf(STDERR, "read_data(): can't open %s\n", argv[1]);
		exit(1);
	}
	if(0) fprintf(STDOUT, "read_data(): DEBUG open %s\n", argv[1]);
	
	a_ct = 0;
	t_ct = 0;
	seq_ct = 0;
	t_seq_ct = 0;
	mod_ct = 1;
	while(fgets(buf, LineWidth, fp_pcd)!=NULL){
		if(0) {
			fprintf(STDOUT,
				"read_data(): DEBUG (a_ct= %d, t_ct= %d) buf= %s \n",
				a_ct, t_ct, buf);
		}
		switch(*(buf+1)) {
			case('P'):	/* <PDB_data ...> */
				atoms =read_pdb_data_line(buf);
				break;
			case('C'):	/* <Coordinates> */
     				/* if(!strncmp("<Coordinates>", buf, 13)) {} */
     				break;
			case('S'):	/* <Slant_tiles ...> */
				s_tiles = read_slant_tiles_line(buf);
				break;
			case('T'): 	/* Tile sequence 1 */
				t_seq_ct = read_tile_sequence_line(buf,
								   t_seq_ct);
				break;
			case('A'): 	/* <Atom_sequence No.=1 ... */
				seq_ct = read_atom_sequence_line(buf, seq_ct,
								 a_ct);
		    		break;
		    	case('a'): 	/* atom : [     1]   ... */
				a_ct = read_atom_line(buf, atoms, a_ct, seq_ct);
				break;
			 
			case('/'):	/* End of ... */
				mod_ct = read_end_line(buf, mod_ct, a_ct,
						       seq_ct, t_ct, s_tiles);
				break;
			case('o'): 	/* Origin = ... */
				read_origine_line(buf, pt_coor);
     				break;
     			case('x'): 	/* x-axis = ... */
				read_x_axis_line(buf, pt_coor);
				break;
	  		case('y'):	/* y-axis = ... */
				read_y_axis_line(buf, pt_coor);
	       			break;
	       		case('z'):	/* z-axis = ... */
				read_z_axis_line(buf, pt_coor);
				break;
		    	case('w'):	/* w-axis = ... */
				read_w_axis_line(buf, pt_coor);
				break;
			case('t'):	/* tile : ... */
				t_ct = read_tile_line(buf, s_tiles, t_ct);
				break;
			case(' '):	/*        center= ... */
				read_space_line(buf, s_tiles, t_ct, pt_coor);
				break;
		}
	}
	if (set_tile_head_and_tail(s_tiles) == FAIL)	exit(1);
	fclose(fp_pcd);
	return s_tiles;
}