/*
 *  pv_trn_table.c
 *  ProteinViewer
 *
 *  Created by Naoto Morikawa on 6/15/06.
 *  Copyright 2006 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pv_main.h"
#include "pv_trn_table.h"

/* extern struct pv_macro my_macro; */
extern struct pe_coor_system  my_coor;
extern struct pe_settings my_settings;

int ston[256];
struct pe_qtos_table qtos[24];
struct pe_trn_table my_trn_table[24];
struct pe_path_of_length_n *npath_table[24];

/* U/D sequence */
/* p_ct =  code[0]*2**(n-1) + code[1]*2**(n-2) + ... + code[n-2]*2 + code[n-1]
*/
void print_npath_table(void)
{
	int i, p_ct, t_ct;
	int n;
	int max_path;
	
	n = (int) my_settings.tiles_per_res;
	max_path = my_settings.max_path;
	
	fprintf(STDOUT, "\n");
	fprintf(STDOUT,
		"N-tile transition table: Dump list starts (max_path= %d)\n",
		max_path);
	if (npath_table[0] == NULL) {
		fprintf(STDOUT,
			"N-tile transition table: Memory not assigned.\n");
		fprintf(STDOUT, "\n");
		return;
	}
	for (i = 0; i < 24; i++) {
		fprintf(STDOUT,
			"npath_table[%2d]:---------------------------------------------------------------\n",
			i);
		for (p_ct = 0; p_ct < max_path; p_ct++) {
			/* U/D sequence */
			fprintf(STDOUT,
				"npath_table[%2d]: code[0]->code[%d] = ",
				i, n);
			for (t_ct = 0; t_ct < n - 1; t_ct++) {
				if (npath_table[i][p_ct].code[t_ct] == C_DOWN)
					fprintf(STDOUT, "D-");
				else
					fprintf(STDOUT, "U-");
			}
			if (npath_table[i][p_ct].code[n - 1] == C_DOWN)
				fprintf(STDOUT, "D\n");
			else
				fprintf(STDOUT, "U\n");
			fprintf(STDOUT,
				"npath_table[%2d]: direction (%8.3f, %8.3f, %8.3f), length= %8.3f\n",
				i,
				npath_table[i][p_ct].c.x - my_trn_table[i].cx,
				npath_table[i][p_ct].c.y - my_trn_table[i].cy,
				npath_table[i][p_ct].c.z - my_trn_table[i].cz,
				npath_table[i][p_ct].c.length);
			
			fprintf(STDOUT,
				"npath_table[%2d]:           from (%8.3f, %8.3f, %8.3f)",
				i, my_trn_table[i].cx, my_trn_table[i].cy,
				my_trn_table[i].cz);
			fprintf(STDOUT, " to (%8.3f, %8.3f, %8.3f)\n",
				npath_table[i][p_ct].c.x,
				npath_table[i][p_ct].c.y,
				npath_table[i][p_ct].c.z);
		}
		fprintf(STDOUT, "\n");
		fflush(STDOUT);
	}
	fprintf(STDOUT, "N-tile transition table: Dump list ends.\n");
	fprintf(STDOUT, "\n");
	return;
}

int npath_code_conversion(int tile_type, int path_ct)
{
	int code, tmp;
	int t_ct;
	
	code = path_ct;
	for (t_ct = my_settings.tiles_per_res - 1; t_ct >= 0; t_ct--) {
		tmp = code % 2;
		if (tmp == 0)
			npath_table[tile_type][path_ct].code[t_ct]= C_DOWN;
		else
			npath_table[tile_type][path_ct].code[t_ct] = C_UP;
		code = code / 2;
	}
	if (DEBUG) {
		if (code != 0) {
			fprintf(STDERR,
				"npath_code_conversioq(): Error conversion failed (code= %d)\n", code);
			return FAIL;
		}
	}
	return SUCCESS;
}

int alloc_npath_code_seq(int tile_type)
{
	int p_ct;
	int tiles_per_res;
	int max_path;
	
	tiles_per_res = (int) my_settings.tiles_per_res;
	max_path = my_settings.max_path;
	for (p_ct = 0; p_ct < max_path; p_ct++) {
		npath_table[tile_type][p_ct].code
		= (PE_CODE *) malloc(tiles_per_res * sizeof(PE_CODE));
		if (npath_table[tile_type][p_ct].code == NULL) {
			fprintf(STDERR, 
				"alloc_npath_code_seq() : ERROR npath_table[%d][%d].code not allocated\n",
				tile_type, p_ct);
			return FAIL;
		}
		if (1 || (my_settings.along_backbone > 0)) {	/* [TODO] */
			npath_table[tile_type][p_ct].bb_c
				= (struct pe_vector *)
				malloc(tiles_per_res *sizeof(struct pe_vector));
			if (npath_table[tile_type][p_ct].bb_c == NULL) {
				fprintf(STDERR, 
					"alloc_npath_code_seq() : ERROR npath_table[%d][%d].bb_c not allocated\n",
					tile_type, p_ct);
				return FAIL;
			}
		}
		if (npath_code_conversion(tile_type, p_ct) == FAIL)
			return FAIL;
	}
	return SUCCESS;
}

/* Count of paths = 2**n
* p_ct =  code[0]*2**(n-1) + code[1]*2**(n-2) + ... + code[n-1]
* where n = tiles_per_res
*/
int alloc_npath_table(void)
{
	int tile_type;
	
	if (npath_table[0] != NULL) {
		fprintf(STDERR, 
			"alloc_npath_code_seq() : ERROR npath_table[0] already allocated\n");
		return FAIL;
	}
	for (tile_type = 0; tile_type < 24; tile_type++) {
		npath_table[tile_type] = (struct pe_path_of_length_n *)
		malloc(my_settings.max_path
		       * sizeof(struct pe_path_of_length_n));
		if (npath_table[tile_type] == NULL) {
			fprintf(STDERR, 
				"alloc_npath_code_seq() : ERROR npath_table[%d] not allocated\n",
				tile_type);
			return FAIL;
		}
		if (alloc_npath_code_seq(tile_type) == FAIL)
			return FAIL;
	}
	return SUCCESS;
}

void compute_position_of_last_tile(int tile_type, int p_ct /* path_ct */)
{
	int cur_type;
	int t_ct;
	float wrk_x, wrk_y, wrk_z;
	
	cur_type = tile_type;
	wrk_x = my_trn_table[cur_type].cx;
	wrk_y = my_trn_table[cur_type].cy;
	wrk_z = my_trn_table[cur_type].cz;
	if (my_settings.along_backbone > 0) {
		npath_table[tile_type][p_ct].bb_c[0].x = wrk_x;
		npath_table[tile_type][p_ct].bb_c[0].y = wrk_y;
		npath_table[tile_type][p_ct].bb_c[0].z = wrk_z;
		npath_table[tile_type][p_ct].bb_c[0].length = 0.0;
	}
	/* The position of the (t_ct+1)-th tile (from 1 to n-1 th) */
	for (t_ct = 0; t_ct < my_settings.tiles_per_res - 1; t_ct++) {
		if (npath_table[tile_type][p_ct].code[t_ct] == C_DOWN) {
			wrk_x = wrk_x + my_trn_table[cur_type].downward_cx
			- my_trn_table[cur_type].cx;
			wrk_y = wrk_y + my_trn_table[cur_type].downward_cy
				- my_trn_table[cur_type].cy;
			wrk_z = wrk_z + my_trn_table[cur_type].downward_cz
				- my_trn_table[cur_type].cz;
			if (npath_table[tile_type][p_ct].code[t_ct+1] == C_DOWN)
				cur_type = 
					my_trn_table[cur_type].downward_D_type;
			else
				cur_type = 
					my_trn_table[cur_type].downward_U_type;
		} else {
			wrk_x = wrk_x + my_trn_table[cur_type].upward_cx
			- my_trn_table[cur_type].cx;
			wrk_y = wrk_y + my_trn_table[cur_type].upward_cy
				- my_trn_table[cur_type].cy;
			wrk_z = wrk_z + my_trn_table[cur_type].upward_cz
				- my_trn_table[cur_type].cz;
			if (npath_table[tile_type][p_ct].code[t_ct+1] == C_DOWN)
				cur_type = my_trn_table[cur_type].upward_D_type;
			else
				cur_type = my_trn_table[cur_type].upward_U_type;
		}
		if (my_settings.along_backbone > 0) {
			npath_table[tile_type][p_ct].bb_c[t_ct+1].x = wrk_x;
			npath_table[tile_type][p_ct].bb_c[t_ct+1].y = wrk_y;
			npath_table[tile_type][p_ct].bb_c[t_ct+1].z = wrk_z;
			npath_table[tile_type][p_ct].bb_c[t_ct+1].length = 0.0;
		}
	}
	/* The position of the n-th tile = the next target atom */
	/* code of (n-1)-th tile = npath_table[tile_type][p_ct].code[t_ct] */
	if (npath_table[tile_type][p_ct].code[t_ct] == C_DOWN) {
		npath_table[tile_type][p_ct].c.x = wrk_x
		+ my_trn_table[cur_type].downward_cx
		- my_trn_table[cur_type].cx;
		npath_table[tile_type][p_ct].c.y = wrk_y
			+ my_trn_table[cur_type].downward_cy
			- my_trn_table[cur_type].cy;
		npath_table[tile_type][p_ct].c.z = wrk_z
			+ my_trn_table[cur_type].downward_cz
			- my_trn_table[cur_type].cz;
	} else {
		npath_table[tile_type][p_ct].c.x = wrk_x
		+ my_trn_table[cur_type].upward_cx
		- my_trn_table[cur_type].cx;
		npath_table[tile_type][p_ct].c.y = wrk_y
			+ my_trn_table[cur_type].upward_cy
			- my_trn_table[cur_type].cy;
		npath_table[tile_type][p_ct].c.z = wrk_z
			+ my_trn_table[cur_type].upward_cz
			- my_trn_table[cur_type].cz;
	}
	/* The type of the last tile */
	npath_table[tile_type][p_ct].type = cur_type;
	wrk_x = npath_table[tile_type][p_ct].c.x - my_trn_table[tile_type].cx;
	wrk_y = npath_table[tile_type][p_ct].c.y - my_trn_table[tile_type].cy;
	wrk_z = npath_table[tile_type][p_ct].c.z - my_trn_table[tile_type].cz;
	npath_table[tile_type][p_ct].c.length
		= sqrt(wrk_x*wrk_x + wrk_y*wrk_y + wrk_z*wrk_z);
	return;
}

/* The adjust vectors */
void adjust_npath_vectors(void)
{
	int n;
	int t_type;
	int p_ct, t_ct;
	float wrk_x, wrk_y, wrk_z;
	
	n = (int) my_settings.tiles_per_res;
	for (t_type = 0; t_type < 24; t_type++) {
		for (p_ct=0; p_ct < my_settings.max_path; p_ct++) {
			wrk_x = npath_table[t_type][p_ct].c.x
				- npath_table[t_type][p_ct].bb_c[0].x;
			wrk_y = npath_table[t_type][p_ct].c.y
				- npath_table[t_type][p_ct].bb_c[0].y;
			wrk_z = npath_table[t_type][p_ct].c.z
				- npath_table[t_type][p_ct].bb_c[0].z;
			npath_table[t_type][p_ct].c.x = wrk_x;
			npath_table[t_type][p_ct].c.y = wrk_y;
			npath_table[t_type][p_ct].c.z = wrk_z;
			npath_table[t_type][p_ct].c.length
				= sqrt(wrk_x*wrk_x + wrk_y*wrk_y + wrk_z*wrk_z);
			for (t_ct = 0; t_ct < n; t_ct++) {
				wrk_x = npath_table[t_type][p_ct].bb_c[t_ct].x
					- npath_table[t_type][p_ct].bb_c[0].x;
				wrk_y = npath_table[t_type][p_ct].bb_c[t_ct].y
					- npath_table[t_type][p_ct].bb_c[0].y;
				wrk_z = npath_table[t_type][p_ct].bb_c[t_ct].z
					- npath_table[t_type][p_ct].bb_c[0].z;
				npath_table[t_type][p_ct].bb_c[t_ct].x = wrk_x;
				npath_table[t_type][p_ct].bb_c[t_ct].y = wrk_y;
				npath_table[t_type][p_ct].bb_c[t_ct].z = wrk_z;
			}
		}
	}
	for (t_type = 0; t_type < 24; t_type++) {
		for (p_ct=0; p_ct < my_settings.max_path; p_ct++) {
			for (t_ct = 0; t_ct < n-1; t_ct++) {
				wrk_x = npath_table[t_type][p_ct].bb_c[t_ct+1].x
				    - npath_table[t_type][p_ct].bb_c[t_ct].x;
				wrk_y = npath_table[t_type][p_ct].bb_c[t_ct+1].y
				    - npath_table[t_type][p_ct].bb_c[t_ct].y;
				wrk_z = npath_table[t_type][p_ct].bb_c[t_ct+1].z
				    - npath_table[t_type][p_ct].bb_c[t_ct].z;
				npath_table[t_type][p_ct].bb_c[t_ct].x = wrk_x;
				npath_table[t_type][p_ct].bb_c[t_ct].y = wrk_y;
				npath_table[t_type][p_ct].bb_c[t_ct].z = wrk_z;
				npath_table[t_type][p_ct].bb_c[t_ct].length
					= sqrt(wrk_x*wrk_x + wrk_y*wrk_y
					       + wrk_z*wrk_z);
			}
			wrk_x = npath_table[t_type][p_ct].c.x
				- npath_table[t_type][p_ct].bb_c[n-1].x;
			wrk_y = npath_table[t_type][p_ct].c.y
				- npath_table[t_type][p_ct].bb_c[n-1].y;
			wrk_z = npath_table[t_type][p_ct].c.z
				- npath_table[t_type][p_ct].bb_c[n-1].z;
			npath_table[t_type][p_ct].bb_c[n-1].x = wrk_x;
			npath_table[t_type][p_ct].bb_c[n-1].y = wrk_y;
			npath_table[t_type][p_ct].bb_c[n-1].z = wrk_z;
			npath_table[t_type][p_ct].bb_c[n-1].length
				= sqrt(wrk_x*wrk_x + wrk_y*wrk_y + wrk_z*wrk_z);
		}
	}
	return;
}

int setup_npath_table(void)
{
	int tile_type;
	int p_ct;
	
	my_settings.max_path = (int) pow(2.0,
					 (double) my_settings.tiles_per_res);
	if (my_settings.max_path <= 0) {
		fprintf(STDERR, "setup_npath_table() :  ERROR max_path = %d\n",
			my_settings.max_path);
		return FAIL;
	}
	if (npath_table[0] == NULL) {
		if (alloc_npath_table() == FAIL) return FAIL;
	}
	for (tile_type = 0; tile_type < 24; tile_type++) {
		for (p_ct = 0; p_ct < my_settings.max_path; p_ct++) {
			compute_position_of_last_tile(tile_type, p_ct);
		}
	}
	if (0) print_npath_table();
	adjust_npath_vectors();
	return SUCCESS;
}

void print_trn_table(void)
{
	int i;
	
	fprintf(STDOUT, "Transition table: Dump list starts\n");
	for (i = 0; i < 24; i++) {
		fprintf(STDOUT,
			"my_trn_table[%2d]:-------------------------------------------------------------------\n",
			i);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (current tile) type= 1%s, center= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].str, my_trn_table[i].cx,
			my_trn_table[i].cy, my_trn_table[i].cz);
		
		fprintf(STDOUT,
			"my_trn_table[%2d]: (downward tile) center= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].downward_cx,
			my_trn_table[i].downward_cy,
			my_trn_table[i].downward_cz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (downward&&Up) type= 1%s, top vertex= (%d, %d, %d, %d) i.e. (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[my_trn_table[i].downward_U_type].str,
			my_trn_table[i].downward_U_top[0],
			my_trn_table[i].downward_U_top[1],
			my_trn_table[i].downward_U_top[2],
			my_trn_table[i].downward_U_top[3],
			my_trn_table[i].downward_U_tx,
			my_trn_table[i].downward_U_ty,
			my_trn_table[i].downward_U_tz);
		fprintf(STDOUT,
			"my_trn_table[%2d]:                center of the next tile= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].downward_U_next_cx,
			my_trn_table[i].downward_U_next_cy,
			my_trn_table[i].downward_U_next_cz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (downward&&Down) type= 1%s, top vertex= (%d, %d, %d, %d) i.e. (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[my_trn_table[i].downward_D_type].str,
			my_trn_table[i].downward_D_top[0],
			my_trn_table[i].downward_D_top[1],
			my_trn_table[i].downward_D_top[2],
			my_trn_table[i].downward_D_top[3],
			my_trn_table[i].downward_D_tx, my_trn_table[i].downward_D_ty,
			my_trn_table[i].downward_D_tz);
		fprintf(STDOUT,
			"my_trn_table[%2d]:                center of the next tile= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].downward_D_next_cx,
			my_trn_table[i].downward_D_next_cy,
			my_trn_table[i].downward_D_next_cz);
		
		fprintf(STDOUT,
			"my_trn_table[%2d]: (upward tile) center= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].upward_cx, my_trn_table[i].upward_cy,
			my_trn_table[i].upward_cz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (upward&&Up) type= 1%s, top vertex= (%d, %d, %d, %d) i.e. (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[my_trn_table[i].upward_U_type].str,
			my_trn_table[i].upward_U_top[0],
			my_trn_table[i].upward_U_top[1],
			my_trn_table[i].upward_U_top[2],
			my_trn_table[i].upward_U_top[3],
			my_trn_table[i].upward_U_tx,
			my_trn_table[i].upward_U_ty,
			my_trn_table[i].upward_U_tz);
		fprintf(STDOUT,
			"my_trn_table[%2d]:                center of the next tile= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].upward_U_next_cx,
			my_trn_table[i].upward_U_next_cy,
			my_trn_table[i].upward_U_next_cz);
		fprintf(STDOUT,
			"my_trn_table[%2d]: (upward&&Down) type= 1%s, top vertex= (%d, %d, %d, %d) i.e. (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[my_trn_table[i].upward_D_type].str,
			my_trn_table[i].upward_D_top[0],
			my_trn_table[i].upward_D_top[1],
			my_trn_table[i].upward_D_top[2],
			my_trn_table[i].upward_D_top[3],
			my_trn_table[i].upward_D_tx,
			my_trn_table[i].upward_D_ty,
			my_trn_table[i].upward_D_tz);
		fprintf(STDOUT,
			"my_trn_table[%2d]:                center of the next tile= (%8.3f, %8.3f, %8.3f)\n",
			i, my_trn_table[i].upward_D_next_cx,
			my_trn_table[i].upward_D_next_cy,
			my_trn_table[i].upward_D_next_cz);
		fflush(STDOUT);
	}			/* for (i=0; i<24; i++) { */
	fprintf(STDOUT, "Transition table: Dump list ends.\n");
	return;
}

/* Set up the table for conversion from strings to integeres. */
void setup_string_to_integer_table(void)
{
	int i;
	
	for (i = 0; i < 256; i++)
		ston[i] = -1;
	
	ston[INDX(XX, YY, ZZ, WW)] = XYZW;
	ston[INDX(XX, ZZ, YY, WW)] = XZYW;
	ston[INDX(YY, XX, ZZ, WW)] = YXZW;
	ston[INDX(YY, ZZ, XX, WW)] = YZXW;
	ston[INDX(ZZ, XX, YY, WW)] = ZXYW;
	ston[INDX(ZZ, YY, XX, WW)] = ZYXW;
	
	ston[INDX(WW, XX, YY, ZZ)] = WXYZ;
	ston[INDX(YY, WW, XX, ZZ)] = YWXZ;
	ston[INDX(WW, YY, XX, ZZ)] = WYXZ;
	ston[INDX(XX, WW, YY, ZZ)] = XWYZ;
	ston[INDX(XX, YY, WW, ZZ)] = XYWZ;
	ston[INDX(YY, XX, WW, ZZ)] = YXWZ;
	
	ston[INDX(ZZ, WW, XX, YY)] = ZWXY;
	ston[INDX(WW, XX, ZZ, YY)] = WXZY;
	ston[INDX(XX, ZZ, WW, YY)] = XZWY;
	ston[INDX(ZZ, XX, WW, YY)] = ZXWY;
	ston[INDX(WW, ZZ, XX, YY)] = WZXY;
	ston[INDX(XX, WW, ZZ, YY)] = XWZY;
	
	ston[INDX(YY, ZZ, WW, XX)] = YZWX;
	ston[INDX(ZZ, YY, WW, XX)] = ZYWX;
	ston[INDX(ZZ, WW, YY, XX)] = ZWYX;
	ston[INDX(WW, YY, ZZ, XX)] = WYZX;
	ston[INDX(YY, WW, ZZ, XX)] = YWZX;
	ston[INDX(WW, ZZ, YY, XX)] = WZYX;
	return;
}

void setup_coordinates_to_tiletype_table(void)
{
	qtos[XYZW].p = XX;
	qtos[XYZW].q = YY;
	qtos[XYZW].r = ZZ;
	qtos[XYZW].s = WW;
	strncpy(qtos[XYZW].str, "[xyz]", 5);
	qtos[XZYW].p = XX;
	qtos[XZYW].q = ZZ;
	qtos[XZYW].r = YY;
	qtos[XZYW].s = WW;
	strncpy(qtos[XZYW].str, "[xzy]", 5);
	qtos[YXZW].p = YY;
	qtos[YXZW].q = XX;
	qtos[YXZW].r = ZZ;
	qtos[YXZW].s = WW;
	strncpy(qtos[YXZW].str, "[yxz]", 5);
	qtos[YZXW].p = YY;
	qtos[YZXW].q = ZZ;
	qtos[YZXW].r = XX;
	qtos[YZXW].s = WW;
	strncpy(qtos[YZXW].str, "[yzx]", 5);
	qtos[ZXYW].p = ZZ;
	qtos[ZXYW].q = XX;
	qtos[ZXYW].r = YY;
	qtos[ZXYW].s = WW;
	strncpy(qtos[ZXYW].str, "[zxy]", 5);
	qtos[ZYXW].p = ZZ;
	qtos[ZYXW].q = YY;
	qtos[ZYXW].r = XX;
	qtos[ZYXW].s = WW;
	strncpy(qtos[ZYXW].str, "[zyx]", 5);
	
	qtos[WXYZ].p = WW;
	qtos[WXYZ].q = XX;
	qtos[WXYZ].r = YY;
	qtos[WXYZ].s = ZZ;
	strncpy(qtos[WXYZ].str, "[wxy]", 5);
	qtos[YWXZ].p = YY;
	qtos[YWXZ].q = WW;
	qtos[YWXZ].r = XX;
	qtos[YWXZ].s = ZZ;
	strncpy(qtos[YWXZ].str, "[ywx]", 5);
	qtos[WYXZ].p = WW;
	qtos[WYXZ].q = YY;
	qtos[WYXZ].r = XX;
	qtos[WYXZ].s = ZZ;
	strncpy(qtos[WYXZ].str, "[wyx]", 5);
	qtos[XWYZ].p = XX;
	qtos[XWYZ].q = WW;
	qtos[XWYZ].r = YY;
	qtos[XWYZ].s = ZZ;
	strncpy(qtos[XWYZ].str, "[xwy]", 5);
	qtos[XYWZ].p = XX;
	qtos[XYWZ].q = YY;
	qtos[XYWZ].r = WW;
	qtos[XYWZ].s = ZZ;
	strncpy(qtos[XYWZ].str, "[xyw]", 5);
	qtos[YXWZ].p = YY;
	qtos[YXWZ].q = XX;
	qtos[YXWZ].r = WW;
	qtos[YXWZ].s = ZZ;
	strncpy(qtos[YXWZ].str, "[yxw]", 5);
	
	qtos[ZWXY].p = ZZ;
	qtos[ZWXY].q = WW;
	qtos[ZWXY].r = XX;
	qtos[ZWXY].s = YY;
	strncpy(qtos[ZWXY].str, "[zwx]", 5);
	qtos[WXZY].p = WW;
	qtos[WXZY].q = XX;
	qtos[WXZY].r = ZZ;
	qtos[WXZY].s = YY;
	strncpy(qtos[WXZY].str, "[wxz]", 5);
	qtos[XZWY].p = XX;
	qtos[XZWY].q = ZZ;
	qtos[XZWY].r = WW;
	qtos[XZWY].s = YY;
	strncpy(qtos[XZWY].str, "[xzw]", 5);
	qtos[ZXWY].p = ZZ;
	qtos[ZXWY].q = XX;
	qtos[ZXWY].r = WW;
	qtos[ZXWY].s = YY;
	strncpy(qtos[ZXWY].str, "[zxw]", 5);
	qtos[WZXY].p = WW;
	qtos[WZXY].q = ZZ;
	qtos[WZXY].r = XX;
	qtos[WZXY].s = YY;
	strncpy(qtos[WZXY].str, "[wzx]", 5);
	qtos[XWZY].p = XX;
	qtos[XWZY].q = WW;
	qtos[XWZY].r = ZZ;
	qtos[XWZY].s = YY;
	strncpy(qtos[XWZY].str, "[xwz]", 5);
	
	qtos[YZWX].p = YY;
	qtos[YZWX].q = ZZ;
	qtos[YZWX].r = WW;
	qtos[YZWX].s = XX;
	strncpy(qtos[YZWX].str, "[yzw]", 5);
	qtos[ZYWX].p = ZZ;
	qtos[ZYWX].q = YY;
	qtos[ZYWX].r = WW;
	qtos[ZYWX].s = XX;
	strncpy(qtos[ZYWX].str, "[zyw]", 5);
	qtos[ZWYX].p = ZZ;
	qtos[ZWYX].q = WW;
	qtos[ZWYX].r = YY;
	qtos[ZWYX].s = XX;
	strncpy(qtos[ZWYX].str, "[zwy]", 5);
	qtos[WYZX].p = WW;
	qtos[WYZX].q = YY;
	qtos[WYZX].r = ZZ;
	qtos[WYZX].s = XX;
	strncpy(qtos[WYZX].str, "[wyz]", 5);
	qtos[YWZX].p = YY;
	qtos[YWZX].q = WW;
	qtos[YWZX].r = ZZ;
	qtos[YWZX].s = XX;
	strncpy(qtos[YWZX].str, "[ywz]", 5);
	qtos[WZYX].p = WW;
	qtos[WZYX].q = ZZ;
	qtos[WZYX].r = YY;
	qtos[WZYX].s = XX;
	strncpy(qtos[WZYX].str, "[wzy]", 5);
	
	if (0) {
		int i;
		
		for (i = 0; i < 24; i++) {
			if (ston[INDX(qtos[i].p, qtos[i].q, qtos[i].r,
				      qtos[i].s)] != i) {
				fprintf(STDERR,
"setup_coordinates_to_tiletype_table() : ERROR  ston-qtos inconsistence\n");
				exit(1);
			}
		}
	}
	return;
}

void set_trn_table_centre_field(int i /* tile_type */)
{
	float VX_x, VX_y, VX_z;
	float VY_x, VY_y, VY_z;
	float VZ_x, VZ_y, VZ_z;
	float VW_x, VW_y, VW_z;
	
	VX_x = my_coor.axes[qtos[i].p].x;
	VX_y = my_coor.axes[qtos[i].p].y;
	VX_z = my_coor.axes[qtos[i].p].z;
	VY_x = my_coor.axes[qtos[i].q].x;
	VY_y = my_coor.axes[qtos[i].q].y;
	VY_z = my_coor.axes[qtos[i].q].z;
	VZ_x = my_coor.axes[qtos[i].r].x;
	VZ_y = my_coor.axes[qtos[i].r].y;
	VZ_z = my_coor.axes[qtos[i].r].z;
	VW_x = my_coor.axes[qtos[i].s].x;
	VW_y = my_coor.axes[qtos[i].s].y;
	VW_z = my_coor.axes[qtos[i].s].z;
	
	my_trn_table[i].cx = (3 * VX_x + 2 * VY_x + VZ_x) / 4;
	my_trn_table[i].cy = (3 * VX_y + 2 * VY_y + VZ_y) / 4;
	my_trn_table[i].cz = (3 * VX_z + 2 * VY_z + VZ_z) / 4;
	my_trn_table[i].bx = VX_x + VY_x + VZ_x;	/* == -VW */
	my_trn_table[i].by = VX_y + VY_y + VZ_y;
	my_trn_table[i].bz = VX_z + VY_z + VZ_z;
	return;
}

void set_trn_table_downward_field(int i /* tile_type */)
{
	float VX_x, VX_y, VX_z;
	float VY_x, VY_y, VY_z;
	float VZ_x, VZ_y, VZ_z;
	float VW_x, VW_y, VW_z;
	
	VX_x = my_coor.axes[qtos[i].p].x;
	VX_y = my_coor.axes[qtos[i].p].y;
	VX_z = my_coor.axes[qtos[i].p].z;
	VY_x = my_coor.axes[qtos[i].q].x;
	VY_y = my_coor.axes[qtos[i].q].y;
	VY_z = my_coor.axes[qtos[i].q].z;
	VZ_x = my_coor.axes[qtos[i].r].x;
	VZ_y = my_coor.axes[qtos[i].r].y;
	VZ_z = my_coor.axes[qtos[i].r].z;
	VW_x = my_coor.axes[qtos[i].s].x;
	VW_y = my_coor.axes[qtos[i].s].y;
	VW_z = my_coor.axes[qtos[i].s].z;
	
	/* |(x/w)[wyz]| */
	my_trn_table[i].downward_cx = (5 * VX_x + 3 * VY_x + 2 * VZ_x) / 4;
	my_trn_table[i].downward_cy = (5 * VX_y + 3 * VY_y + 2 * VZ_y) / 4;
	my_trn_table[i].downward_cz = (5 * VX_z + 3 * VY_z + 2 * VZ_z) / 4;
	/* |(x/w)[wyx]| */
	my_trn_table[i].downward_U_next_cx = (5 * VX_x + 2 * VY_x - VW_x) / 4;
	my_trn_table[i].downward_U_next_cy = (5 * VX_y + 2 * VY_y - VW_y) / 4;
	my_trn_table[i].downward_U_next_cz = (5 * VX_z + 2 * VY_z - VW_z) / 4;
	/* x/w */
	my_trn_table[i].downward_U_tx = VX_x - VW_x;
	my_trn_table[i].downward_U_ty = VX_y - VW_y;
	my_trn_table[i].downward_U_tz = VX_z - VW_z;
	/* x/w */
	my_trn_table[i].downward_U_top[qtos[i].p] = 1;
	my_trn_table[i].downward_U_top[qtos[i].q] = 0;
	my_trn_table[i].downward_U_top[qtos[i].r] = 0;
	my_trn_table[i].downward_U_top[qtos[i].s] = -1;
	/* [wyz] */
	my_trn_table[i].downward_U_type
		= ston[INDX(qtos[i].s, qtos[i].q, qtos[i].r, qtos[i].p)];
	/* |xy[zxy]| */
	my_trn_table[i].downward_D_next_cx
		= (6 * VX_x + 5 * VY_x + 3 * VZ_x) / 4;
	my_trn_table[i].downward_D_next_cy = (6*VX_y + 5*VY_y + 3*VZ_y) / 4;
	my_trn_table[i].downward_D_next_cz = (6*VX_z + 5*VY_z + 3*VZ_z) / 4;
	/* x */
	my_trn_table[i].downward_D_tx = VX_x;
	my_trn_table[i].downward_D_ty = VX_y;
	my_trn_table[i].downward_D_tz = VX_z;
	my_trn_table[i].downward_D_top[qtos[i].p] = 1;
	my_trn_table[i].downward_D_top[qtos[i].q] = 0;
	my_trn_table[i].downward_D_top[qtos[i].r] = 0;
	my_trn_table[i].downward_D_top[qtos[i].s] = 0;
	/* [yzx] */
	my_trn_table[i].downward_D_type = ston[INDX(qtos[i].q, qtos[i].r, qtos[i].p, qtos[i].s)];
	return;
}

void set_trn_table_upward_field(int i /* tile_type */)
{
	float VX_x, VX_y, VX_z;
	float VY_x, VY_y, VY_z;
	float VZ_x, VZ_y, VZ_z;
	float VW_x, VW_y, VW_z;
	
	VX_x = my_coor.axes[qtos[i].p].x;
	VX_y = my_coor.axes[qtos[i].p].y;
	VX_z = my_coor.axes[qtos[i].p].z;
	VY_x = my_coor.axes[qtos[i].q].x;
	VY_y = my_coor.axes[qtos[i].q].y;
	VY_z = my_coor.axes[qtos[i].q].z;
	VZ_x = my_coor.axes[qtos[i].r].x;
	VZ_y = my_coor.axes[qtos[i].r].y;
	VZ_z = my_coor.axes[qtos[i].r].z;
	VW_x = my_coor.axes[qtos[i].s].x;
	VW_y = my_coor.axes[qtos[i].s].y;
	VW_z = my_coor.axes[qtos[i].s].z;
	
	my_trn_table[i].upward_cx = (3 * VX_x + 2 * VY_x + VW_x) / 4;
	my_trn_table[i].upward_cy = (3 * VX_y + 2 * VY_y + VW_y) / 4;
	my_trn_table[i].upward_cz = (3 * VX_z + 2 * VY_z + VW_z) / 4;
	/* (1/z)[zxy] */
	my_trn_table[i].upward_U_next_cx = (2 * VX_x - VZ_x + VW_x) / 4;
	my_trn_table[i].upward_U_next_cy = (2 * VX_y - VZ_y + VW_y) / 4;
	my_trn_table[i].upward_U_next_cz = (2 * VX_z - VZ_z + VW_z) / 4;
	/* 1/z */
	my_trn_table[i].upward_U_tx = -VZ_x;
	my_trn_table[i].upward_U_ty = -VZ_y;
	my_trn_table[i].upward_U_tz = -VZ_z;
	my_trn_table[i].upward_U_top[qtos[i].p] = 0;
	my_trn_table[i].upward_U_top[qtos[i].q] = 0;
	my_trn_table[i].upward_U_top[qtos[i].r] = -1;
	my_trn_table[i].upward_U_top[qtos[i].s] = 0;
	/* [zxy] */
	my_trn_table[i].upward_U_type
		= ston[INDX(qtos[i].r, qtos[i].p, qtos[i].q, qtos[i].s)];
	/* |x[ywx]| */
	my_trn_table[i].upward_D_next_cx = (5*VX_x + 3*VY_x + 2*VW_x) / 4;
	my_trn_table[i].upward_D_next_cy = (5*VX_y + 3*VY_y + 2*VW_y) / 4;
	my_trn_table[i].upward_D_next_cz = (5*VX_z + 3*VY_z + 2*VW_z) / 4;
	/* 1 */
	my_trn_table[i].upward_D_tx = 0.0;
	my_trn_table[i].upward_D_ty = 0.0;
	my_trn_table[i].upward_D_tz = 0.0;
	my_trn_table[i].upward_D_top[qtos[i].p] = 0;
	my_trn_table[i].upward_D_top[qtos[i].q] = 0;
	my_trn_table[i].upward_D_top[qtos[i].r] = 0;
	my_trn_table[i].upward_D_top[qtos[i].s] = 0;
	/* [xyw] */
	my_trn_table[i].upward_D_type
		= ston[INDX(qtos[i].p, qtos[i].q, qtos[i].s, qtos[i].r)];
	return;
}

void check_trn_table(float ave_dis)
{
	int flag;
	int i; /* tile_type */
	int nxt_type;
	float wrk_float;
	
	flag = 0;
	for (i = 0; i < 24; i++) {
		/*downward && U */
		nxt_type = my_trn_table[i].downward_U_type;
		wrk_float = my_trn_table[i].downward_cx
			- my_trn_table[i].downward_U_tx;
		wrk_float = my_trn_table[nxt_type].cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 1;
		wrk_float = my_trn_table[i].downward_cy
			- my_trn_table[i].downward_U_ty;
		wrk_float = my_trn_table[nxt_type].cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 2;
		wrk_float = my_trn_table[i].downward_cz
			- my_trn_table[i].downward_U_tz;
		wrk_float = my_trn_table[nxt_type].cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 4;
		wrk_float = my_trn_table[i].downward_U_next_cx
			- my_trn_table[i].downward_U_tx;
		wrk_float = my_trn_table[nxt_type].upward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 10;
		wrk_float = my_trn_table[i].downward_U_next_cy
			- my_trn_table[i].downward_U_ty;
		wrk_float = my_trn_table[nxt_type].upward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 20;
		wrk_float = my_trn_table[i].downward_U_next_cz
			- my_trn_table[i].downward_U_tz;
		wrk_float = my_trn_table[nxt_type].upward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 40;
		wrk_float = my_trn_table[i].cx - my_trn_table[i].downward_U_tx;
		wrk_float = my_trn_table[nxt_type]. downward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 100;
		wrk_float = my_trn_table[i].cy - my_trn_table[i].downward_U_ty;
		wrk_float = my_trn_table[nxt_type].downward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 200;
		wrk_float = my_trn_table[i].cz - my_trn_table[i].downward_U_tz;
		wrk_float = my_trn_table[nxt_type].downward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 400;
		wrk_float = my_trn_table[i].upward_cx
			- my_trn_table[i].downward_U_tx;
		wrk_float = my_trn_table[nxt_type].downward_U_next_cx
			- wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 1000;
		wrk_float = my_trn_table[i].upward_cy
			- my_trn_table[i].downward_U_ty;
		wrk_float = my_trn_table[nxt_type].downward_U_next_cy
			- wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 2000;
		wrk_float = my_trn_table[i].upward_cz
			- my_trn_table[i].downward_U_tz;
		wrk_float
		= my_trn_table[nxt_type].downward_U_next_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 4000;
		if (my_trn_table[nxt_type].downward_U_type != i)
			flag = flag + 10000;
		if (flag > 0) {
			fprintf(STDERR,
"check_trn_table() : ERROR  - my_trn_table[%d], downward&&U (flag=%d) -\n",
				i, flag);
			exit(1);
		}
		
		/*downward && D */
		nxt_type = my_trn_table[i].downward_D_type;
		wrk_float = my_trn_table[i].downward_cx
			- my_trn_table[i].downward_D_tx;
		wrk_float = my_trn_table[nxt_type].cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 1;
		wrk_float = my_trn_table[i].downward_cy
			- my_trn_table[i].downward_D_ty;
		wrk_float = my_trn_table[nxt_type].cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 2;
		wrk_float = my_trn_table[i].downward_cz
			- my_trn_table[i].downward_D_tz;
		wrk_float = my_trn_table[nxt_type].cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 4;
		wrk_float = my_trn_table[i].downward_D_next_cx
			- my_trn_table[i].downward_D_tx;
		wrk_float = my_trn_table[nxt_type].downward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 10;
		wrk_float = my_trn_table[i].downward_D_next_cy
			- my_trn_table[i].downward_D_ty;
		wrk_float = my_trn_table[nxt_type].downward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 20;
		wrk_float = my_trn_table[i].downward_D_next_cz -
			my_trn_table[i].downward_D_tz;
		wrk_float = my_trn_table[nxt_type].downward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 40;
		wrk_float = my_trn_table[i].cx - my_trn_table[i].downward_D_tx;
		wrk_float = my_trn_table[nxt_type].upward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 100;
		wrk_float = my_trn_table[i].cy - my_trn_table[i].downward_D_ty;
		wrk_float = my_trn_table[nxt_type].upward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 200;
		wrk_float = my_trn_table[i].cz - my_trn_table[i].downward_D_tz;
		wrk_float = my_trn_table[nxt_type].upward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 400;
		wrk_float = my_trn_table[i].upward_cx
			- my_trn_table[i].downward_D_tx;
		wrk_float = my_trn_table[nxt_type].upward_U_next_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 1000;
		wrk_float = my_trn_table[i].upward_cy
			- my_trn_table[i].downward_D_ty;
		wrk_float = my_trn_table[nxt_type].upward_U_next_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 2000;
		wrk_float = my_trn_table[i].upward_cz
			- my_trn_table[i].downward_D_tz;
		wrk_float = my_trn_table[nxt_type].upward_U_next_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 4000;
		if (my_trn_table[nxt_type]. upward_U_type != i)
			flag = flag + 10000;
		if (flag > 0) {
			fprintf(STDERR,
"check_trn_table() : ERROR   my_trn_table[%d], downward&&D (flag=%d) -\n",
				i, flag);
			exit(1);
		}
		
		/*upward && U */
		nxt_type = my_trn_table[i].upward_U_type;
		wrk_float = my_trn_table[i].upward_cx
			- my_trn_table[i].upward_U_tx;
		wrk_float = my_trn_table[nxt_type].cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 1;
		wrk_float = my_trn_table[i].upward_cy
			- my_trn_table[i].upward_U_ty;
		wrk_float = my_trn_table[nxt_type].cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 2;
		wrk_float = my_trn_table[i].upward_cz
			- my_trn_table[i].upward_U_tz;
		wrk_float = my_trn_table[nxt_type].cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 4;
		wrk_float = my_trn_table[i].upward_U_next_cx
			- my_trn_table[i].upward_U_tx;
		wrk_float = my_trn_table[nxt_type].upward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 10;
		wrk_float = my_trn_table[i].upward_U_next_cy
			- my_trn_table[i].upward_U_ty;
		wrk_float = my_trn_table[nxt_type].upward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 20;
		wrk_float = my_trn_table[i].upward_U_next_cz
			- my_trn_table[i].upward_U_tz;
		wrk_float = my_trn_table[nxt_type].upward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 40;
		wrk_float = my_trn_table[i].cx - my_trn_table[i].upward_U_tx;
		wrk_float = my_trn_table[nxt_type].downward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 100;
		wrk_float = my_trn_table[i].cy - my_trn_table[i].upward_U_ty;
		wrk_float = my_trn_table[nxt_type].downward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 200;
		wrk_float = my_trn_table[i].cz - my_trn_table[i].upward_U_tz;
		wrk_float = my_trn_table[nxt_type].downward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 400;
		wrk_float = my_trn_table[i].downward_cx
			- my_trn_table[i].upward_U_tx;
		wrk_float = my_trn_table[nxt_type].downward_D_next_cx
			- wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 1000;
		wrk_float = my_trn_table[i].downward_cy
			- my_trn_table[i].upward_U_ty;
		wrk_float = my_trn_table[nxt_type].downward_D_next_cy
			- wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 2000;
		wrk_float = my_trn_table[i].downward_cz
			- my_trn_table[i].upward_U_tz;
		wrk_float = my_trn_table[nxt_type].downward_D_next_cz
			- wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 4000;
		if (my_trn_table[nxt_type].downward_D_type != i)
			flag = flag + 10000;
		if (flag > 0) {
			fprintf(STDERR,
"check_trn_table() : ERROR  - my_trn_table[%d], upward&&U (flag=%d) -\n",
				i, flag);
			exit(1);
		}
		
		/*upward && D */
		nxt_type = my_trn_table[i].upward_D_type;
		wrk_float = my_trn_table[i].upward_cx
			- my_trn_table[i].upward_D_tx;
		wrk_float = my_trn_table[nxt_type].cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 1;
		wrk_float = my_trn_table[i].upward_cy
			- my_trn_table[i].upward_D_ty;
		wrk_float = my_trn_table[nxt_type].cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 2;
		wrk_float = my_trn_table[i].upward_cz
			- my_trn_table[i].upward_D_tz;
		wrk_float = my_trn_table[nxt_type].cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 4;
		wrk_float = my_trn_table[i].upward_D_next_cx
			- my_trn_table[i].upward_D_tx;
		wrk_float = my_trn_table[nxt_type].downward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 10;
		wrk_float = my_trn_table[i].upward_D_next_cy
			- my_trn_table[i].upward_D_ty;
		wrk_float = my_trn_table[nxt_type].downward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 20;
		wrk_float = my_trn_table[i].upward_D_next_cz -
			my_trn_table[i].upward_D_tz;
		wrk_float = my_trn_table[nxt_type].downward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 40;
		wrk_float = my_trn_table[i].cx - my_trn_table[i].upward_D_tx;
		wrk_float = my_trn_table[nxt_type].upward_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 100;
		wrk_float = my_trn_table[i].cy - my_trn_table[i].upward_D_ty;
		wrk_float = my_trn_table[nxt_type].upward_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 200;
		wrk_float = my_trn_table[i].cz - my_trn_table[i].upward_D_tz;
		wrk_float = my_trn_table[nxt_type].upward_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 400;
		wrk_float = my_trn_table[i].downward_cx
			- my_trn_table[i].upward_D_tx;
		wrk_float = my_trn_table[nxt_type].upward_D_next_cx - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 1000;
		wrk_float = my_trn_table[i].downward_cy
			- my_trn_table[i].upward_D_ty;
		wrk_float = my_trn_table[nxt_type].upward_D_next_cy - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 2000;
		wrk_float = my_trn_table[i].downward_cz
			- my_trn_table[i].upward_D_tz;
		wrk_float = my_trn_table[nxt_type].upward_D_next_cz - wrk_float;
		if (FLOAT_NEQ(wrk_float/ave_dis))
			flag = flag + 4000;
		if (my_trn_table[nxt_type].upward_D_type != i) {
			flag = flag + 10000;
		}
		if (flag > 0) {
			fprintf(STDERR,
"check_trn_table() : ERROR   my_trn_table[%d], upward&&D (flag=%d)\n",
				i, flag);
			exit(1);
		}		
	}
	return;
}

void setup_transition_table(float ave_dis)
{
	int tile_type;

	setup_string_to_integer_table();	/* ston[] */
	setup_coordinates_to_tiletype_table();	/* qtos[] */

	/* Set up the transition table */
	for (tile_type = 0; tile_type < 24; tile_type++) {
		strcpy(my_trn_table[tile_type].str, qtos[tile_type].str);
		set_trn_table_centre_field(tile_type);
		set_trn_table_downward_field(tile_type);
		set_trn_table_upward_field(tile_type);
	}
	if (0) check_trn_table(ave_dis);

	return;
}
